"""This script is a helper utility to run entire pipelines from a single script.

To do this effectively, this runner functionally takes a "meta configuration" file that contains:

  1. The path to the pipeline configuration file.
  2. Configuration details for how to run each stage of the pipeline, including mappings to the underlying
     stage scripts and Hydra launcher configurations for each stage to control parallelism, resources, etc.
"""

import logging
import subprocess
from pathlib import Path

import hydra
import yaml
from omegaconf import DictConfig, OmegaConf

from .configs import RUNNER_CONFIG_YAML, PipelineConfig
from .configs.utils import OmegaConfResolver

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

logger = logging.getLogger(__name__)


def get_parallelization_args(
    parallelization_cfg: dict | DictConfig | None, default_parallelization_cfg: dict | DictConfig
) -> list[str]:
    """Extracts the specific parallelization arguments given the default and stage-specific configurations.

    Args:
        parallelization_cfg: The stage-specific parallelization configuration.
        default_parallelization_cfg: The default parallelization configuration.

    Returns:
        A list of command-line arguments for parallelization.

    Examples:
        >>> get_parallelization_args({}, {})
        []
        >>> get_parallelization_args(None, {"n_workers": 4})
        []
        >>> get_parallelization_args({"launcher": "joblib"}, {})
        ['--multirun', 'worker="range(0,1)"', 'hydra/launcher=joblib']
        >>> get_parallelization_args({"n_workers": 2, "launcher_params": "foo"}, {})
        Traceback (most recent call last):
            ...
        ValueError: If launcher_params is provided, launcher must also be provided.
        >>> get_parallelization_args({"n_workers": 2}, {})
        ['--multirun', 'worker="range(0,2)"']
        >>> get_parallelization_args({"launcher": "slurm"}, {"n_workers": 3, "launcher": "joblib"})
        ['--multirun', 'worker="range(0,3)"', 'hydra/launcher=slurm']
        >>> get_parallelization_args(
        ...     {"n_workers": 2, "launcher": "joblib"},
        ...     {"n_workers": 5, "launcher_params": {"foo": "bar"}},
        ... )
        ['--multirun', 'worker="range(0,2)"', 'hydra/launcher=joblib', 'hydra.launcher.foo=bar']
        >>> get_parallelization_args(
        ...     {"n_workers": 5, "launcher_params": {"biz": "baz"}, "launcher": "slurm"}, {}
        ... )
        ['--multirun', 'worker="range(0,5)"', 'hydra/launcher=slurm', 'hydra.launcher.biz=baz']
    """

    if parallelization_cfg is None:
        return []

    if len(parallelization_cfg) == 0 and len(default_parallelization_cfg) == 0:
        return []

    if "n_workers" in parallelization_cfg:
        n_workers = parallelization_cfg["n_workers"]
    elif "n_workers" in default_parallelization_cfg:
        n_workers = default_parallelization_cfg["n_workers"]
    else:
        n_workers = 1

    parallelization_args = [
        "--multirun",
        f'worker="range(0,{n_workers})"',
    ]

    if "launcher" in parallelization_cfg:
        launcher = parallelization_cfg["launcher"]
    elif "launcher" in default_parallelization_cfg:
        launcher = default_parallelization_cfg["launcher"]
    else:
        launcher = None

    if launcher is None:
        if "launcher_params" in parallelization_cfg:
            raise ValueError("If launcher_params is provided, launcher must also be provided.")

        return parallelization_args

    parallelization_args.append(f"hydra/launcher={launcher}")

    if "launcher_params" in parallelization_cfg:
        launcher_params = parallelization_cfg["launcher_params"]
    elif "launcher_params" in default_parallelization_cfg:
        launcher_params = default_parallelization_cfg["launcher_params"]
    else:
        launcher_params = {}

    for k, v in launcher_params.items():
        parallelization_args.append(f"hydra.launcher.{k}={v}")

    return parallelization_args


def run_stage(
    cfg: DictConfig,
    pipeline_cfg: PipelineConfig,
    stage_name: str,
    default_parallelization_cfg: dict | DictConfig | None = None,
    runner_fn: callable = subprocess.run,  # For dependency injection
):
    """Runs a single stage of the pipeline.

    Args:
        cfg: The configuration for the entire pipeline.
        stage_name: The name of the stage to run.

    Raises:
        ValueError: If the stage fails to run.

    Examples:
        >>> def fake_shell_succeed(cmd, shell, capture_output):
        ...     print(cmd)
        ...     return subprocess.CompletedProcess(args=cmd, returncode=0, stdout=b"", stderr=b"")
        >>> def fake_shell_fail(cmd, shell, capture_output):
        ...     print(cmd)
        ...     return subprocess.CompletedProcess(args=cmd, returncode=1, stdout=b"", stderr=b"")
        >>> cfg = DictConfig(
        ...     {
        ...         "pipeline_config_fp": "pipeline_config.yaml",
        ...         "do_profile": False,
        ...         "_stage_runners": {
        ...             "reshard_to_split": {"_script": "not used"},
        ...             "fit_vocabulary_indices": {},
        ...             "reorder_measurements": {"script": "baz_script"},
        ...         },
        ...     }
        ... )
        >>> pipeline_cfg = PipelineConfig(
        ...     stages=[
        ...         "reshard_to_split",
        ...         {"fit_vocabulary_indices": {"_script": "foobar"}},
        ...         "reorder_measurements",
        ...     ],
        ... )
        >>> run_stage(cfg, pipeline_cfg, "reshard_to_split", runner_fn=fake_shell_succeed)
        MEDS_transform-stage pipeline_config.yaml reshard_to_split stage=reshard_to_split
        >>> run_stage(cfg, pipeline_cfg, "fit_vocabulary_indices", runner_fn=fake_shell_succeed)
        foobar stage=fit_vocabulary_indices
        >>> run_stage(cfg, pipeline_cfg, "reorder_measurements", runner_fn=fake_shell_succeed)
        baz_script stage=reorder_measurements
        >>> cfg.do_profile = True
        >>> run_stage(cfg, pipeline_cfg, "reorder_measurements", runner_fn=fake_shell_succeed)
        baz_script stage=reorder_measurements
            ++hydra.callbacks.profiler._target_=hydra_profiler.profiler.ProfilerCallback
        >>> cfg._stage_runners.reorder_measurements.parallelize = {"n_workers": 2}
        >>> cfg.do_profile = False
        >>> run_stage(cfg, pipeline_cfg, "reorder_measurements", runner_fn=fake_shell_succeed)
        baz_script --multirun stage=reorder_measurements worker="range(0,2)"
        >>> run_stage(cfg, pipeline_cfg, "reorder_measurements", runner_fn=fake_shell_fail)
        Traceback (most recent call last):
            ...
        ValueError: Stage reorder_measurements failed via ...

        Errors are also raised if the stage runner is improperly configured.

        >>> cfg = DictConfig(
        ...     {
        ...         "pipeline_config_fp": "pipeline_config.yaml",
        ...         "do_profile": False,
        ...         "_stage_runners": {
        ...             "reshard_to_split": {"_base_stage": "belongs in the stage"},
        ...         },
        ...     }
        ... )
        >>> pipeline_cfg = PipelineConfig(stages=["reshard_to_split"])
        >>> run_stage(cfg, pipeline_cfg, "reshard_to_split", runner_fn=fake_shell_succeed)
        Traceback (most recent call last):
            ...
        ValueError: Put _base_stage args is in your pipeline config
    """

    if default_parallelization_cfg is None:
        default_parallelization_cfg = {}

    do_profile = cfg.get("do_profile", False)
    stage_config = pipeline_cfg.parsed_stages_by_name[stage_name].config
    stage_runner_config = cfg._stage_runners.get(stage_name, {})

    script = None
    if "script" in stage_runner_config:
        script = stage_runner_config.get("script")
    elif "_script" in stage_config:
        script = stage_config.get("_script")
    elif "_base_stage" in stage_runner_config:
        raise ValueError("Put _base_stage args is in your pipeline config")
    else:
        script = f"MEDS_transform-stage {cfg.pipeline_config_fp} {stage_name}"

    command_parts = [
        script,
        f"stage={stage_name}",
    ]

    parallelization_args = get_parallelization_args(
        stage_runner_config.get("parallelize", {}), default_parallelization_cfg
    )

    if parallelization_args:
        multirun = parallelization_args.pop(0)
        command_parts = command_parts[:1] + [multirun] + command_parts[1:] + parallelization_args

    if do_profile:
        command_parts.append("++hydra.callbacks.profiler._target_=hydra_profiler.profiler.ProfilerCallback")

    full_cmd = " ".join(command_parts)
    logger.info(f"Running command: {full_cmd}")
    command_out = runner_fn(full_cmd, shell=True, capture_output=True)

    # https://stackoverflow.com/questions/21953835/run-subprocess-and-print-output-to-logging

    stderr = command_out.stderr.decode()
    stdout = command_out.stdout.decode()
    logger.info(f"Command output:\n{stdout}")
    logger.info(f"Command error:\n{stderr}")

    if command_out.returncode != 0:
        raise ValueError(
            f"Stage {stage_name} failed via {full_cmd} with return code {command_out.returncode}."
        )


@hydra.main(
    version_base=None, config_path=str(RUNNER_CONFIG_YAML.parent), config_name=RUNNER_CONFIG_YAML.stem
)
def main(cfg: DictConfig):
    """Runs the entire pipeline, end-to-end, based on the configuration provided.

    This script will launch many subsidiary commands via `subprocess`, one for each stage of the specified
    pipeline.

    Args:
        cfg: The Hydra configuration for the entire pipeline. See `configs/_runner.yaml` for details.

    Raises:
        ValueError: If the pipeline configuration is invalid, if any stage fails to run, or if you try to run
            with `do_profile` without the `hydra-profiler` package installed.

    Examples:
        >>> cfg = DictConfig(
        ...     {
        ...         "pipeline_config_fp": "???",  # we'll fill this in each test.
        ...         "log_dir": "???",  # We'll fill this in each test; it is added by Hydra during CLI usage.
        ...         "do_profile": False,
        ...         "_stage_runners": {},
        ...     }
        ... )
        >>> pipeline_cfg = DictConfig(
        ...     {
        ...         "stages": [
        ...             "reshard_to_split",
        ...             {"count_codes": {"_base_stage": "aggregate_code_metadata"}},
        ...         ],
        ...     }
        ... )

    We'll consistently mock out the `run_stage` calls to avoid running the actual stages; instead, when
    called, we'll just print out the stages. To do this, we'll set up a helper to run the full test.

        >>> def mock_run_stage(cfg, pipeline_cfg, stage, *args, **kwargs):
        ...     if "default_parallelization_cfg" in kwargs and kwargs["default_parallelization_cfg"]:
        ...         print(f"Running {stage} with: {kwargs['default_parallelization_cfg']}")
        ...     else:
        ...         print(f"Running {stage}")
        >>> def profile_main(test_dir: Path):
        ...     # Copy the cfg
        ...     test_cfg = DictConfig(cfg)
        ...     # Make th log dir
        ...     log_dir = test_dir / "logs"
        ...     log_dir.mkdir(exist_ok=True, parents=True)
        ...     test_cfg.log_dir = str(log_dir)
        ...     # Save the pipeline config
        ...     pipeline_cfg_fp = test_dir / "pipeline_config.yaml"
        ...     OmegaConf.save(pipeline_cfg, pipeline_cfg_fp)
        ...     test_cfg.pipeline_config_fp = str(pipeline_cfg_fp)
        ...     # Mock the run_stage function
        ...     with patch("MEDS_transforms.runner.run_stage") as mock:
        ...         mock.side_effect = mock_run_stage
        ...         # Run the main function
        ...         main(test_cfg)

    Now we can run some tests. Let's start with the runner in normal operation. We see it "runs" each stage
    (here just prints them):

        >>> with tempfile.TemporaryDirectory() as test_dir:
        ...     test_dir = Path(test_dir)
        ...     profile_main(test_dir)
        Running reshard_to_split
        Running count_codes

    What controls exist? Firstly, upon completion of each stage, it also writes a ".done" file to the log dir
    for that stage, and a global "_all_stages.done" file when the pipeline is done. Let's see them:

        >>> with tempfile.TemporaryDirectory() as test_dir:
        ...     test_dir = Path(test_dir)
        ...     profile_main(test_dir)
        ...     print("Log contents:")
        ...     print_directory(test_dir / "logs")
        Running reshard_to_split
        Running count_codes
        Log contents:
        ├── _all_stages.done
        ├── count_codes.done
        └── reshard_to_split.done

    If the global done file exists in advance, the pipeline will skip all stages:

        >>> with tempfile.TemporaryDirectory() as test_dir:
        ...     test_dir = Path(test_dir)
        ...     log_dir = test_dir / "logs"
        ...     log_dir.mkdir()
        ...     global_done_file = log_dir / "_all_stages.done"
        ...     global_done_file.touch()
        ...     profile_main(test_dir)
        ...     print("Log contents:")
        ...     print_directory(log_dir)
        Log contents:
        └── _all_stages.done

    This applies to each individual stage's done file too:

        >>> with tempfile.TemporaryDirectory() as test_dir:
        ...     test_dir = Path(test_dir)
        ...     log_dir = test_dir / "logs"
        ...     log_dir.mkdir()
        ...     reshard_done_file = log_dir / "reshard_to_split.done"
        ...     reshard_done_file.touch()
        ...     profile_main(test_dir)
        ...     print("Log contents:")
        ...     print_directory(log_dir)
        Running count_codes
        Log contents:
        ├── _all_stages.done
        ├── count_codes.done
        └── reshard_to_split.done

    We can also control parallelization, either through the pipeline config or the stage runner config:

        >>> cfg.parallelize = {"n_workers": 2}
        >>> with tempfile.TemporaryDirectory() as test_dir:
        ...     profile_main(Path(test_dir))
        Running reshard_to_split with: {'n_workers': 2}
        Running count_codes with: {'n_workers': 2}

    Stage runner parallelization takes priority, though:

        >>> cfg._stage_runners.parallelize = {"n_workers": 3}
        >>> with tempfile.TemporaryDirectory() as test_dir:
        ...     profile_main(Path(test_dir))
        Running reshard_to_split with: {'n_workers': 3}
        Running count_codes with: {'n_workers': 3}

    The runner will throw an error if the pipeline speified has no stages:

        >>> pipeline_cfg.stages = []
        >>> with tempfile.TemporaryDirectory() as test_dir:
        ...     profile_main(Path(test_dir))
        Traceback (most recent call last):
            ...
        ValueError: Pipeline configuration must specify at least one stage.
    """

    pipeline_config = PipelineConfig.from_arg(cfg.pipeline_config_fp)

    stages = [s.name for s in pipeline_config.parsed_stages]
    if not stages:
        raise ValueError("Pipeline configuration must specify at least one stage.")

    log_dir = Path(cfg.log_dir)

    if cfg.get("do_profile", False):  # pragma: no cover
        try:
            import hydra_profiler  # noqa: F401
        except ImportError as e:
            raise ValueError(
                "You can't run in profiling mode without installing hydra-profiler. Try installing "
                "MEDS-transforms with the 'profiler' optional dependency: "
                "`pip install MEDS-transforms[profiler]`."
            ) from e

    global_done_file = log_dir / "_all_stages.done"
    if global_done_file.exists():
        logger.info("All stages are already complete. Exiting.")
        return

    if "parallelize" in cfg._stage_runners:
        default_parallelization_cfg = cfg._stage_runners.parallelize
    elif "parallelize" in cfg:
        default_parallelization_cfg = cfg.parallelize
    else:
        default_parallelization_cfg = None

    for stage in stages:
        done_file = log_dir / f"{stage}.done"

        if done_file.exists():
            logger.info(f"Skipping stage {stage} as it is already complete.")
        else:
            logger.info(f"Running stage: {stage}")
            run_stage(cfg, pipeline_config, stage, default_parallelization_cfg=default_parallelization_cfg)
            done_file.touch()

    global_done_file.touch()


@OmegaConfResolver
def load_yaml_file(path: str | None) -> dict | DictConfig:
    """Loads a YAML file as an OmegaConf object.

    Args:
        path: The path to the YAML file.

    Returns:
        The OmegaConf object representing the YAML file, or None if no path is provided.

    Raises:
        FileNotFoundError: If the file does not exist.

    Examples:
        >>> load_yaml_file(None)
        {}
        >>> load_yaml_file("nonexistent_file.yaml")
        Traceback (most recent call last):
            ...
        FileNotFoundError: File nonexistent_file.yaml does not exist.
        >>> with tempfile.NamedTemporaryFile(suffix=".yaml") as f:
        ...     _ = f.write(b"foo: bar")
        ...     f.flush()
        ...     load_yaml_file(f.name)
        {'foo': 'bar'}
        >>> with tempfile.NamedTemporaryFile(suffix=".yaml") as f:
        ...     cfg = OmegaConf.create({"foo": "bar"})
        ...     OmegaConf.save(cfg, f.name)
        ...     load_yaml_file(f.name)
        {'foo': 'bar'}
    """

    if not path:
        return {}

    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"File {path} does not exist.")

    try:
        return OmegaConf.load(path)
    except Exception as e:  # pragma: no cover
        logger.warning(f"Failed to load {path} as an OmegaConf: {e}. Trying as a plain YAML file.")
        yaml_text = path.read_text()
        return yaml.load(yaml_text, Loader=Loader)



from importlib.resources import files

from .. import __package_name__
from .dataset import DatasetConfig
from .pipeline import PipelineConfig
from .stage import StageConfig

RUNNER_CONFIG_YAML = files(__package_name__).joinpath("configs/_runner.yaml")

__all__ = ["DatasetConfig", "PipelineConfig", "StageConfig"]



"""Basic utilities for parallelizable mapreduces on sharded MEDS datasets with caching and locking."""

import hashlib
import logging
from collections.abc import Callable
from datetime import UTC, datetime
from pathlib import Path
from typing import NamedTuple

from omegaconf import DictConfig

logger = logging.getLogger(__name__)


class InOutFilePair(NamedTuple):
    in_fp: Path
    out_fp: Path


SHARD_ITR_FNTR_T = Callable[[DictConfig, str, str], tuple[list[InOutFilePair], bool]]


def shuffle_shards(shards: list[str], cfg: DictConfig) -> list[str]:
    """Shuffle the shards in a deterministic, pseudo-random way based on the worker ID in the configuration.

    Args:
        shards: The list of shards to shuffle.
        cfg: The configuration dictionary for the overall pipeline. Should (possibly) contain the following
            keys (some are optional, as marked below):
            - `worker` (optional): The worker ID for the MR worker; this is also used to seed the
              randomization process. If not provided, the randomization process will be unseeded.

    Returns:
        The shuffled list of shards.

    Examples:
        >>> shards = ["train/0", "train/1", "tuning", "held_out"]
        >>> shuffle_shards(shards, DictConfig({"worker": 1}))
        ['train/1', 'held_out', 'tuning', 'train/0']
        >>> shuffle_shards(shards, DictConfig({"worker": 2}))
        ['tuning', 'held_out', 'train/1', 'train/0']

        It can also shuffle the shards without a worker ID, but the order is then based on the time, which
        is not consistent across runs.

        >>> sorted(shuffle_shards(shards, DictConfig({})))
        ['held_out', 'train/0', 'train/1', 'tuning']

        If the shards aren't unique, it will error

        >>> shards = ["train/0", "train/0", "tuning", "held_out"]
        >>> shuffle_shards(shards, DictConfig({"worker": 1}))
        Traceback (most recent call last):
            ...
        ValueError: Shards must be unique, but found duplicates: train/0
    """

    if len(shards) != len(set(shards)):
        duplicates = sorted({shard for shard in shards if shards.count(shard) > 1})
        raise ValueError(f"Shards must be unique, but found duplicates: {', '.join(duplicates)}")

    add_str = str(cfg.get("worker", datetime.now(tz=UTC)))

    def hash_fn(shard: str) -> int:
        return int(hashlib.sha256((add_str + shard).encode("utf-8")).hexdigest(), 16)

    return sorted(shards, key=hash_fn)


def shard_iterator(
    cfg: DictConfig,
    out_suffix: str = ".parquet",
    in_prefix: str = "",
) -> tuple[list[InOutFilePair], bool]:
    """Returns a list of the shards found in the input directory and their corresponding output directories.

    Args:
        cfg: The configuration dictionary for the overall pipeline. Should (possibly) contain the following
            keys (some are optional, as marked below):
            - `stage_cfg.data_input_dir` (mandatory): The directory containing the input data.
            - `stage_cfg.output_dir` (mandatory): The directory to write the output data.
            - `stage_cfg.train_only` (optional): The prefix of the shards to process (e.g.,
              `"train/"`). If not provided, all shards will be processed.
            - `worker` (optional): The worker ID for the MR worker; this is also used to seed the
              randomization process. If not provided, the randomization process will be unseeded.
        out_suffix: The suffix of the output files. Defaults to ".parquet".
        in_prefix: The prefix of the input files. Defaults to "". This must be a full path component. It can
            end with a slash but even if it doesn't it will be interpreted as a full path component.

    Yields:
        Randomly shuffled pairs of input and output file paths for each shard. The randomization process is
        seeded by the worker ID in ``cfg``, if provided, otherwise it is left unseeded.

    Examples:
        >>> df = pl.DataFrame(
        ...     {
        ...         "subject_id": [1, 2, 3, 4, 5, 6, 7, 8, 9],
        ...         "code": ["A", "B", "C", "D", "E", "F", "G", "H", "I"],
        ...         "time": [1, 2, 3, 4, 5, 6, 1, 2, 3],
        ...     }
        ... )
        >>> shards = {"train/0": [1, 2, 3, 4], "train/1": [5, 6, 7], "tuning": [8], "held_out": [9]}
        >>> def write_dfs(
        ...     input_dir: Path, df: pl.DataFrame = df, shards: dict = shards, sfx: str = ".parquet"
        ... ):
        ...     for shard_name, subject_ids in shards.items():
        ...         df = df.filter(pl.col("subject_id").is_in(subject_ids))
        ...         shard_fp = input_dir / f"{shard_name}{sfx}"
        ...         shard_fp.parent.mkdir(exist_ok=True, parents=True)
        ...         if sfx == ".parquet":
        ...             df.write_parquet(shard_fp)
        ...         elif sfx == ".csv":
        ...             df.write_csv(shard_fp)
        ...         else:
        ...             raise ValueError(f"Unsupported suffix {sfx}")
        ...     return

    By default, this will load all shards in the input directory and write specify their appropriate output
    directories:

        >>> with tempfile.TemporaryDirectory() as tmp:
        ...     root = Path(tmp)
        ...     input_dir = root / "data"
        ...     output_dir = root / "output"
        ...     write_dfs(input_dir)
        ...     cfg = DictConfig(
        ...         {
        ...             "stage_cfg": {"data_input_dir": str(input_dir), "output_dir": str(output_dir)},
        ...             "worker": 1,
        ...         }
        ...     )
        ...     fps, includes_only_train = shard_iterator(cfg)
        >>> [(i.relative_to(root), o.relative_to(root)) for i, o in fps]
        [(PosixPath('data/train/1.parquet'),  PosixPath('output/train/1.parquet')),
         (PosixPath('data/held_out.parquet'), PosixPath('output/held_out.parquet')),
         (PosixPath('data/tuning.parquet'),   PosixPath('output/tuning.parquet')),
         (PosixPath('data/train/0.parquet'),  PosixPath('output/train/0.parquet'))]
        >>> includes_only_train
        False

    Different workers will shuffle the shards differently:

        >>> with tempfile.TemporaryDirectory() as tmp:
        ...     root = Path(tmp)
        ...     input_dir = root / "data"
        ...     output_dir = root / "output"
        ...     write_dfs(input_dir)
        ...     cfg = DictConfig(
        ...         {
        ...             "stage_cfg": {"data_input_dir": str(input_dir), "output_dir": str(output_dir)},
        ...             "worker": 2,
        ...         }
        ...     )
        ...     fps, includes_only_train = shard_iterator(cfg)
        >>> [(i.relative_to(root), o.relative_to(root)) for i, o in fps]
        [(PosixPath('data/tuning.parquet'),   PosixPath('output/tuning.parquet')),
         (PosixPath('data/held_out.parquet'), PosixPath('output/held_out.parquet')),
         (PosixPath('data/train/1.parquet'),  PosixPath('output/train/1.parquet')),
         (PosixPath('data/train/0.parquet'),  PosixPath('output/train/0.parquet'))]
        >>> includes_only_train
        False

    We can also make it look within a specific input subdir of the data directory and change the output
    suffix. Note that using a specific input subdir is _different_ than requesting it load only train.

        >>> with tempfile.TemporaryDirectory() as tmp:
        ...     root = Path(tmp)
        ...     input_dir = root / "data"
        ...     output_dir = root / "output"
        ...     write_dfs(input_dir)
        ...     cfg = DictConfig(
        ...         {
        ...             "stage_cfg": {"data_input_dir": str(input_dir), "output_dir": str(output_dir)},
        ...             "worker": 1,
        ...         }
        ...     )
        ...     fps, includes_only_train = shard_iterator(cfg, in_prefix="train", out_suffix=".csv")
        >>> [(i.relative_to(root), o.relative_to(root)) for i, o in fps]
        [(PosixPath('data/train/0.parquet'),  PosixPath('output/0.csv')),
         (PosixPath('data/train/1.parquet'),  PosixPath('output/1.csv'))]
        >>> includes_only_train
        False

    We can also make it load only 'train' shards, in the case that there are shards with a valid "train/"
    prefix.

        >>> with tempfile.TemporaryDirectory() as tmp:
        ...     root = Path(tmp)
        ...     input_dir = root / "data"
        ...     output_dir = root / "output"
        ...     write_dfs(input_dir)
        ...     cfg = DictConfig(
        ...         {
        ...             "stage_cfg": {
        ...                 "data_input_dir": str(input_dir),
        ...                 "output_dir": str(output_dir),
        ...                 "train_only": True,
        ...             },
        ...             "worker": 1,
        ...         }
        ...     )
        ...     fps, includes_only_train = shard_iterator(cfg)
        >>> [(i.relative_to(root), o.relative_to(root)) for i, o in fps]
        [(PosixPath('data/train/1.parquet'),  PosixPath('output/train/1.parquet')),
         (PosixPath('data/train/0.parquet'),  PosixPath('output/train/0.parquet'))]
        >>> includes_only_train
        True

    The train prefix used is precisely `train/` -- other uses of train will not work:

        >>> wrong_pfx_shards = {"train": [1, 2, 3], "train_1": [4, 5, 6], "train-2": [7, 8, 9]}
        >>> with tempfile.TemporaryDirectory() as tmp:
        ...     root = Path(tmp)
        ...     input_dir = root / "data"
        ...     output_dir = root / "output"
        ...     write_dfs(input_dir, shards=wrong_pfx_shards)
        ...     cfg = DictConfig(
        ...         {
        ...             "stage_cfg": {
        ...                 "data_input_dir": str(input_dir),
        ...                 "output_dir": str(output_dir),
        ...                 "train_only": True,
        ...             },
        ...             "worker": 1,
        ...         }
        ...     )
        ...     fps, includes_only_train = shard_iterator(cfg)
        >>> [(i.relative_to(root), o.relative_to(root)) for i, o in fps]
        [(PosixPath('data/train_1.parquet'),  PosixPath('output/train_1.parquet')),
         (PosixPath('data/train-2.parquet'),  PosixPath('output/train-2.parquet')),
         (PosixPath('data/train.parquet'),  PosixPath('output/train.parquet'))]
        >>> includes_only_train
        False

    If there are no such shards, then it loads them all and assumes the filtering will be handled via the
    splits parquet file.

        >>> no_pfx_shards = {"0": [1, 2, 3], "1": [4, 5, 6], "2": [7, 8, 9]}
        >>> with tempfile.TemporaryDirectory() as tmp:
        ...     root = Path(tmp)
        ...     input_dir = root / "data"
        ...     output_dir = root / "output"
        ...     write_dfs(input_dir, shards=no_pfx_shards)
        ...     cfg = DictConfig(
        ...         {
        ...             "stage_cfg": {
        ...                 "data_input_dir": str(input_dir),
        ...                 "output_dir": str(output_dir),
        ...                 "train_only": True,
        ...             },
        ...             "worker": 1,
        ...         }
        ...     )
        ...     fps, includes_only_train = shard_iterator(cfg)
        >>> [(i.relative_to(root), o.relative_to(root)) for i, o in fps]
        [(PosixPath('data/0.parquet'), PosixPath('output/0.parquet')),
         (PosixPath('data/1.parquet'), PosixPath('output/1.parquet')),
         (PosixPath('data/2.parquet'), PosixPath('output/2.parquet'))]
        >>> includes_only_train
        False

    If it can't find any files, it will error:

        >>> fps, includes_only_train = shard_iterator(cfg)
        Traceback (most recent call last):
            ...
        FileNotFoundError: No shards found in ... with suffix .parquet. Directory contents:...
    """

    input_dir = Path(cfg.stage_cfg.data_input_dir)
    output_dir = Path(cfg.stage_cfg.output_dir)

    in_suffix = ".parquet"

    if in_prefix:
        input_dir = input_dir / in_prefix

    shards = []
    for p in input_dir.glob(f"**/*{in_suffix}"):
        relative_path = p.relative_to(input_dir)
        shard_name = str(relative_path)
        shard_name = shard_name[: -len(in_suffix)]
        shards.append(shard_name)

    if not shards:
        raise FileNotFoundError(
            f"No shards found in {input_dir} with suffix {in_suffix}. Directory contents: "
            f"{', '.join(str(p.relative_to(input_dir)) for p in input_dir.glob('**/*'))}"
        )

    # We initialize this to False and overwrite it if we find dedicated train shards.
    includes_only_train = False

    train_only = cfg.stage_cfg.get("train_only", None)
    train_shards = [shard_name for shard_name in shards if shard_name.startswith("train/")]
    if train_only and train_shards:
        shards = train_shards
        includes_only_train = True
    elif train_only:
        logger.warning(
            f"train_only={train_only} requested but no dedicated train shards found; processing all shards "
            "and relying on `subject_splits.parquet` for filtering."
        )

    shards = shuffle_shards(shards, cfg)

    logger.info(f"Mapping computation over a maximum of {len(shards)} shards")

    out = []
    for sp in shards:
        in_fp = input_dir / f"{sp}{in_suffix}"
        out_fp = output_dir / f"{sp}{out_suffix}"
        # TODO: Could add checking logic for existence of in_fp and/or out_fp here.
        out.append(InOutFilePair(in_fp, out_fp))

    return out, includes_only_train



import re
import subprocess

SCRIPT_TEMPLATE = "MEDS_transform-stage {pipeline} {stage_name}"


def test_stage_entry_point_help():
    result = subprocess.run("MEDS_transform-stage", check=False, shell=True, capture_output=True)
    assert result.returncode != 0

    help_str = result.stdout.decode()
    assert "Usage: " in help_str and "Available stages:" in help_str

    result = subprocess.run("MEDS_transform-stage --help", check=False, shell=True, capture_output=True)
    assert result.returncode == 0
    assert result.stdout.decode() == help_str

    result = subprocess.run("MEDS_transform-stage foo", check=False, shell=True, capture_output=True)
    assert result.returncode != 0
    assert result.stdout.decode() == help_str


def test_stage_entry_point_errors():
    for pipeline, stage, want_err in [
        ("not_real.yaml", "occlude_outliers", "Pipeline YAML file 'not_real.yaml' does not exist."),
        ("__null__", "not_real_stage", "Stage 'not_real_stage' not registered"),
        (
            "pkg://non_existent_pkg.file.yaml",
            "occlude_outliers",
            re.compile("Package 'non_existent_pkg' not found"),
        ),
    ]:
        script = SCRIPT_TEMPLATE.format(pipeline=pipeline, stage_name=stage)
        result = subprocess.run(script, check=False, shell=True, capture_output=True)
        assert result.returncode != 0
        if isinstance(want_err, str):
            assert want_err in result.stderr.decode()
        else:
            assert re.search(want_err, result.stderr.decode()) is not None



"""Tests a multi-stage pre-processing pipeline via the Runner utility. Only checks final outputs.

In this test, the following stages are run:
  - filter_subjects
  - add_time_derived_measurements
  - fit_outlier_detection
  - occlude_outliers
  - fit_normalization
  - fit_vocabulary_indices
  - normalization

The stage configuration arguments will be as given in the yaml block below:
"""

import subprocess
import tempfile
from pathlib import Path

import pytest

from MEDS_transforms.pytest_plugin import pipeline_tester

RUNNER_SCRIPT = "MEDS_transform-pipeline"

PIPELINE_YAML = """
input_dir: {input_dir}
output_dir: {output_dir}

description: "A test pipeline for the MEDS-transforms pipeline runner."

stages:
  - filter_subjects:
      min_events_per_subject: 5
  - add_time_derived_measurements:
      age:
        DOB_code: "DOB"
        age_code: "AGE"
        age_unit: "years"
      time_of_day:
        time_of_day_code: "TIME_OF_DAY"
        endpoints: [6, 12, 18, 24]
  - fit_outlier_detection:
      _base_stage: "aggregate_code_metadata"
      aggregations:
        - "values/n_occurrences"
        - "values/sum"
        - "values/sum_sqd"
  - occlude_outliers:
      stddev_cutoff: 1
  - fit_normalization:
      _base_stage: "aggregate_code_metadata"
      aggregations:
        - "code/n_occurrences"
        - "code/n_subjects"
        - "values/n_occurrences"
        - "values/sum"
        - "values/sum_sqd"
  - fit_vocabulary_indices
  - normalization
"""

STAGE_SCENARIO_SEQUENCE = [
    "filter_subjects",
    "add_time_derived_measurements/in_example_pipeline",
    "aggregate_code_metadata/in_example_pipeline/fit_outlier_detection",
    "occlude_outliers/in_example_pipeline",
    "aggregate_code_metadata/in_example_pipeline/fit_normalization",
    "fit_vocabulary_indices/in_example_pipeline",
    "normalization/in_example_pipeline",
]


# Normally, you wouldn't need to specify all of these scripts, but in testing with local scripts we need to
# specify them all as they need to point to their python paths.
PARALLEL_STAGE_RUNNER_YAML = """
parallelize:
  n_workers: 2
  launcher: "joblib"
"""


def test_example_pipeline():
    pipeline_tester(PIPELINE_YAML, None, STAGE_SCENARIO_SEQUENCE)


@pytest.mark.parallelized
def test_example_pipeline_parallel():
    pipeline_tester(PIPELINE_YAML, PARALLEL_STAGE_RUNNER_YAML, STAGE_SCENARIO_SEQUENCE)


NO_ARGS_HELP_STR = """
== MEDS-Transforms Pipeline Runner ==
MEDS-Transforms Pipeline Runner is a command line tool for running entire MEDS-transform pipelines in a single
command.

**MEDS-transforms Pipeline description:**

No description provided.
"""

WITH_CONFIG_HELP_STR = """
== MEDS-Transforms Pipeline Runner ==
MEDS-Transforms Pipeline Runner is a command line tool for running entire MEDS-transform pipelines in a single
command.

**MEDS-transforms Pipeline description:**

A test pipeline for the MEDS-transforms pipeline runner.
"""


def test_pipeline_help():
    out = subprocess.run(f"{RUNNER_SCRIPT} -h", shell=True, check=True, capture_output=True)
    assert NO_ARGS_HELP_STR.strip() == out.stdout.decode("utf-8").strip()

    with tempfile.TemporaryDirectory() as tmpdir:
        input_dir = Path(tmpdir) / "input"
        output_dir = Path(tmpdir) / "output"

        pipeline_str = PIPELINE_YAML.format(input_dir=input_dir, output_dir=output_dir)

        pipeline_fp = Path(tmpdir) / "pipeline.yaml"
        pipeline_fp.write_text(pipeline_str)

        out = subprocess.run(
            f"{RUNNER_SCRIPT} -h pipeline_config_fp={pipeline_fp}",
            shell=True,
            check=True,
            capture_output=True,
        )
        assert WITH_CONFIG_HELP_STR.strip() == out.stdout.decode("utf-8").strip()



"""This file contains code to aid in defining and programmatically loading stage examples.

This code is not an example of how to use the package, but rather code to help operationalize packaging and
exposing examples for stages within this and derived packages programmatically for testing.
"""

from __future__ import annotations

import subprocess
import tempfile
import textwrap
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar

import polars as pl
from meds import DatasetMetadataSchema, code_metadata_filepath
from meds_testing_helpers.dataset import MEDSDataset
from meds_testing_helpers.static_sample_data import SIMPLE_STATIC_SHARDED_BY_SPLIT
from omegaconf import DictConfig, OmegaConf
from polars.testing import assert_frame_equal
from pretty_print_directory import list_directory
from yaml import load as load_yaml

if TYPE_CHECKING:
    from collections.abc import Callable

_SPACE = "    "
_BRANCH = "│   "

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader


def dict_to_hydra_kwargs(d: dict[str, str]) -> list[str]:
    """Converts a dictionary to a hydra kwargs string for testing purposes.

    Args:
        d: The dictionary to convert.

    Returns:
        A string representation of the dictionary in hydra kwargs (dot-list) format.

    Raises:
        ValueError: If a key in the dictionary is not dot-list compatible.

    Examples:
        >>> args = dict_to_hydra_kwargs(
        ...     {"a": 1, "b": "foo", "c": {"d": True, "e": False, "f": ["foo", "bar"], "g": None}}
        ... )
        >>> for arg in args:
        ...     print(arg)
        a=1
        b=foo
        c.d=true
        c.e=false
        'c.f=["foo", "bar"]'
        ~c.g
        >>> from datetime import datetime
        >>> dict_to_hydra_kwargs({"a": 1, 2: "foo"})
        Traceback (most recent call last):
            ...
        ValueError: Expected all keys to be strings, got 2
        >>> dict_to_hydra_kwargs({"a": datetime(2021, 11, 1)})
        Traceback (most recent call last):
            ...
        ValueError: Unexpected type for value for key a: <class 'datetime.datetime'>: 2021-11-01 00:00:00
    """

    modifier_chars = ["~", "'", "++", "+"]

    out = []
    for k, v in d.items():
        if not isinstance(k, str):
            raise ValueError(f"Expected all keys to be strings, got {k}")
        match v:
            case bool() if v is True:
                out.append(f"{k}=true")
            case bool() if v is False:
                out.append(f"{k}=false")
            case None:
                out.append(f"~{k}")
            case str() | int() | float():
                out.append(f"{k}={v}")
            case dict():
                inner_kwargs = dict_to_hydra_kwargs(v)
                for inner_kv in inner_kwargs:
                    handled = False
                    for mod in modifier_chars:
                        if inner_kv.startswith(mod):
                            out.append(f"{mod}{k}.{inner_kv[len(mod) :]}")
                            handled = True
                            break
                    if not handled:
                        out.append(f"{k}.{inner_kv}")
            case list() | tuple():
                v = list(v)
                v_str_inner = ", ".join(f'"{x}"' for x in v)
                out.append(f"'{k}=[{v_str_inner}]'")
            case _:
                raise ValueError(f"Unexpected type for value for key {k}: {type(v)}: {v}")

    return out


def read_metadata_only(fp: Path, **schema_updates) -> pl.DataFrame:
    """Reads a code metadata file from a yaml representation.

    Eventually, this should be replaced by functionality in `meds_testing_helpers`. TODO(mmd): File an issue
    there to track.

    Args:
        fp: The path to the yaml file.
        schema_updates: Optional schema updates to apply.

    Returns:
        A Polars DataFrame containing the metadata.

    Raises:
        ValueError: If the yaml file does not contain the expected structure or the metadata encoding type is
            not supported.

    Examples:
        >>> metadata_df = pl.DataFrame(
        ...     {
        ...         "code": ["foo", "bar"],
        ...         "description": ["Foo", "Bar"],
        ...         "fake_number": [1, 2],
        ...     }
        ... )
        >>> csv_rep = {"metadata/codes.parquet": metadata_df.write_csv()}
        >>> rows_rep = {"metadata/codes.parquet": metadata_df.to_dicts()}
        >>> with tempfile.NamedTemporaryFile(suffix=".yaml") as tmp:
        ...     yaml_fp = Path(tmp.name)
        ...     OmegaConf.save(csv_rep, yaml_fp)
        ...     read_metadata_only(Path(tmp.name))
        shape: (2, 3)
        ┌──────┬─────────────┬─────────────┐
        │ code ┆ description ┆ fake_number │
        │ ---  ┆ ---         ┆ ---         │
        │ str  ┆ str         ┆ i64         │
        ╞══════╪═════════════╪═════════════╡
        │ foo  ┆ Foo         ┆ 1           │
        │ bar  ┆ Bar         ┆ 2           │
        └──────┴─────────────┴─────────────┘
        >>> with tempfile.NamedTemporaryFile(suffix=".yaml") as tmp:
        ...     yaml_fp = Path(tmp.name)
        ...     OmegaConf.save(csv_rep, yaml_fp)
        ...     read_metadata_only(Path(tmp.name), fake_number=pl.Int32)
        shape: (2, 3)
        ┌──────┬─────────────┬─────────────┐
        │ code ┆ description ┆ fake_number │
        │ ---  ┆ ---         ┆ ---         │
        │ str  ┆ str         ┆ i32         │
        ╞══════╪═════════════╪═════════════╡
        │ foo  ┆ Foo         ┆ 1           │
        │ bar  ┆ Bar         ┆ 2           │
        └──────┴─────────────┴─────────────┘
        >>> with tempfile.NamedTemporaryFile(suffix=".yaml") as tmp:
        ...     yaml_fp = Path(tmp.name)
        ...     OmegaConf.save(rows_rep, yaml_fp)
        ...     read_metadata_only(Path(tmp.name))
        shape: (2, 3)
        ┌──────┬─────────────┬─────────────┐
        │ code ┆ description ┆ fake_number │
        │ ---  ┆ ---         ┆ ---         │
        │ str  ┆ str         ┆ i64         │
        ╞══════╪═════════════╪═════════════╡
        │ foo  ┆ Foo         ┆ 1           │
        │ bar  ┆ Bar         ┆ 2           │
        └──────┴─────────────┴─────────────┘
        >>> with tempfile.NamedTemporaryFile(suffix=".yaml") as tmp:
        ...     yaml_fp = Path(tmp.name)
        ...     OmegaConf.save(rows_rep, yaml_fp)
        ...     read_metadata_only(Path(tmp.name), fake_number=pl.String)
        shape: (2, 3)
        ┌──────┬─────────────┬─────────────┐
        │ code ┆ description ┆ fake_number │
        │ ---  ┆ ---         ┆ ---         │
        │ str  ┆ str         ┆ str         │
        ╞══════╪═════════════╪═════════════╡
        │ foo  ┆ Foo         ┆ 1           │
        │ bar  ┆ Bar         ┆ 2           │
        └──────┴─────────────┴─────────────┘

        Errors are raised if the yaml file is not a filepath:

        >>> read_metadata_only("foo")
        Traceback (most recent call last):
            ...
        TypeError: Expected a Path object, got <class 'str'>: foo

        Or if the file does not exist:

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     read_metadata_only(Path(tmpdir) / "not_real.yaml")
        Traceback (most recent call last):
            ...
        FileNotFoundError: File /tmp/tmp.../not_real.yaml does not exist.

        Or if the yaml file isn't in the right structure

        >>> with tempfile.NamedTemporaryFile(suffix=".yaml") as tmp:
        ...     yaml_fp = Path(tmp.name)
        ...     OmegaConf.save(["foo", "bar", "baz"], yaml_fp)
        ...     read_metadata_only(Path(tmp.name))
        Traceback (most recent call last):
            ...
        ValueError: Expected YAML file to contain 'metadata/codes.parquet: ' pointing to contents, but got:
            ['foo', 'bar', 'baz']

        Or the data in the metadata key isn't in the right format

        >>> cols_rep = {"metadata/codes.parquet": metadata_df.to_dict(as_series=False)}
        >>> with tempfile.NamedTemporaryFile(suffix=".yaml") as tmp:
        ...     yaml_fp = Path(tmp.name)
        ...     OmegaConf.save(cols_rep, yaml_fp)
        ...     read_metadata_only(Path(tmp.name))
        Traceback (most recent call last):
            ...
        ValueError: Unsupported data type for metadata: <class 'dict'>
    """

    if not isinstance(fp, Path):
        raise TypeError(f"Expected a Path object, got {type(fp)}: {fp}")
    if not fp.is_file():
        raise FileNotFoundError(f"File {fp} does not exist.")

    data = load_yaml(fp.read_text(), Loader=Loader)

    if not isinstance(data, dict) or len(data) != 1 or next(iter(data.keys())) != code_metadata_filepath:
        raise ValueError(
            f"Expected YAML file to contain '{code_metadata_filepath}: ' pointing to contents, "
            f"but got:\n{data}"
        )

    key = next(iter(data.keys()))

    val = data[key]

    match data[key]:
        case str() as csv:
            return MEDSDataset.parse_csv(csv, **schema_updates)
        case list() as rows:
            return pl.from_dicts(rows, schema_overrides=schema_updates)
        case _:
            raise ValueError(f"Unsupported data type for metadata: {type(val)}")


@dataclass
class TestEnv:
    """A dataclass to encapsulate the test environment for a stage.

    This is largely just useful for type safety and for pretty printing during debugging and testing.
    """

    script: str
    output_dir: Path
    test_dir: Path
    input_dir: Path
    config_yaml_fp: Path | None = None

    def __str__(self) -> str:
        lines = [f"Test Environment in {self.test_dir}", "  - Files:"]
        lines.extend(list_directory(self.test_dir, prefix=_SPACE))
        lines.append(f"  - Input sub-directory: {self.input_dir.relative_to(self.test_dir)}")
        lines.append(f"  - Output sub-directory: {self.output_dir.relative_to(self.test_dir)}")

        if self.config_yaml_fp:
            lines.append(f"  - Config yaml file: {self.config_yaml_fp.relative_to(self.test_dir)}")
            cfg_yaml_contents = self.config_yaml_fp.read_text().strip()
            lines.extend(textwrap.indent(cfg_yaml_contents, _SPACE + _BRANCH).splitlines())
        lines.append(f"  - Script: {self.script}")
        return "\n".join(lines)


@dataclass
class StageExample:
    """A dataclass to encapsulate an example of a stage being used.

    This is used in the automated stage testing infrastructure both for testing the built-in MEDS transforms
    stages and for downstream package stages.

    Eventually, it may be used for documenting stages as well -- if this would be useful to you, please file a
    GitHub issue to request it so we can track interest.

    Attributes:
        stage_name: The name of the stage.
        scenario_name: The name of the scenario. If the string "." is used, it is replaced with `None`.
        stage_cfg: The configuration options for this example -- e.g., if the stage is run, with these
            options, on the specified in_data, the output should match the want_data or want_metadata.
        want_data: The expected output data for the stage. If `None`, then want_metadata must be provided and
            the stage is a metadata stage.
        want_metadata: The expected output metadata for the stage. If `None`, then want_data must be provided
            and the stage is a data stage.
        in_data: The input data for the stage. If `None`, then the default static data
            (`meds_testing_helpers.static_sample_data.SIMPLE_STATIC_SHARDED_BY_SPLIT`) is used.
        do_use_config_yaml: Whether to use a config.yaml file for the stage. If `True`, then the stage is run
            only with the parameters set via a file, not the command line. Defaults to `False`.

    Raises:
        ValueError: If neither want_data nor want_metadata is provided, or if both are provided.

    Examples:
        >>> metadata_df = pl.DataFrame({"code": ["foo", "bar"], "description": ["Foo", "Bar"]})
        >>> example = StageExample(stage_name="example_stage", want_metadata=metadata_df)
        >>> print(example)
        StageExample [example_stage]
          stage_cfg: {}
          want_metadata:
            shape: (2, 2)
            ┌──────┬─────────────┐
            │ code ┆ description │
            │ ---  ┆ ---         │
            │ str  ┆ str         │
            ╞══════╪═════════════╡
            │ foo  ┆ Foo         │
            │ bar  ┆ Bar         │
            └──────┴─────────────┘

    An example must have at least `want_data` or `want_metadata` provided, and cannot have both:

        >>> StageExample(stage_name="nothing_given")
        Traceback (most recent call last):
            ...
        ValueError: Either want_data or want_metadata must be provided.
        >>> StageExample(
        ...     stage_name="both_given",
        ...     want_data=MEDSDataset(root_dir=simple_static_MEDS),
        ...     want_metadata=pl.DataFrame(),
        ... )
        Traceback (most recent call last):
            ...
        ValueError: Either want_data or want_metadata must be provided, but not both.

    It has some helpful properties that you can access too, for its name, test kwargs, and testing usage:

        >>> metadata_df = pl.DataFrame({"code": ["foo", "bar"], "description": ["Foo", "Bar"]})
        >>> example = StageExample(stage_name="example_stage", want_metadata=metadata_df)
        >>> print(example.full_name)
        example_stage
        >>> print(example.do_use_config_yaml)
        False
        >>> print(example.cmd_pipeline_cfg)
        None
        >>> print(example.cmd_args)
        []
        >>> print(example._err_prefix)
        Stage example example_stage Failed:

    Note that the scenario name is ignored if it is set to "." as well as if it is not set at all. This is to
    reflect the fact that a singleton example directory will have a relative path to root of "." and should be
    inferred to have a null scenario name.

        >>> example = StageExample(stage_name="example_stage", scenario_name=".", want_metadata=metadata_df)
        >>> print(example.scenario_name)
        None

    You can also create an example with a scenario name, stage configuration arguments, test kwargs, and
    output (want) data instead of metadata:

        >>> data_df = pl.DataFrame({"subject_id": [1], "time": [1], "code": ["A"], "numeric_value": [None]})
        >>> data = MEDSDataset(data_shards={"0": data_df}, dataset_metadata=DatasetMetadataSchema())
        >>> example_data = StageExample(
        ...     stage_name="with_scenario",
        ...     scenario_name="example_scenario",
        ...     stage_cfg={"arg1": "value1", "arg2": "value2"},
        ...     want_data=data,
        ...     do_use_config_yaml=True,
        ... )
        >>> print(example_data)
        StageExample [with_scenario/example_scenario]
          stage_cfg: {'arg1': 'value1', 'arg2': 'value2'}
          do_use_config_yaml: True
          want_data:
            MEDSDataset:
            dataset_metadata:
            data_shards:
              - 0:
                pyarrow.Table
                subject_id: int64
                time: timestamp[us]
                code: string
                numeric_value: float
                ----
                subject_id: [[1]]
                time: [[1970-01-01 00:00:00.000001]]
                code: [["A"]]
                numeric_value: [[null]]
            code_metadata:
              pyarrow.Table
              code: string
              description: string
              parent_codes: list<item: string>
                child 0, item: string
              ----
              code: []
              description: []
              parent_codes: []
            subject_splits: None

    You can also set an input dataset for the stage example manually as well; if it is unset (as above), it
    defaults to `meds_testing_helpers.static_sample_data.SIMPLE_STATIC_SHARDED_BY_SPLIT` in automated tests:

        >>> example_with_in_data = StageExample(
        ...     stage_name="with_scenario",
        ...     scenario_name="example_scenario",
        ...     stage_cfg={"arg1": "value1", "arg2": "value2"},
        ...     want_metadata=metadata_df,
        ...     in_data=data,
        ...     do_use_config_yaml=True,
        ... )
        >>> print(example_with_in_data)
        StageExample [with_scenario/example_scenario]
          stage_cfg: {'arg1': 'value1', 'arg2': 'value2'}
          do_use_config_yaml: True
          in_data:
            MEDSDataset:
            dataset_metadata:
            data_shards:
              - 0:
                pyarrow.Table
                subject_id: int64
                time: timestamp[us]
                code: string
                numeric_value: float
                ----
                subject_id: [[1]]
                time: [[1970-01-01 00:00:00.000001]]
                code: [["A"]]
                numeric_value: [[null]]
            code_metadata:
              pyarrow.Table
              code: string
              description: string
              parent_codes: list<item: string>
                child 0, item: string
              ----
              code: []
              description: []
              parent_codes: []
            subject_splits: None
          want_metadata:
            shape: (2, 2)
            ┌──────┬─────────────┐
            │ code ┆ description │
            │ ---  ┆ ---         │
            │ str  ┆ str         │
            ╞══════╪═════════════╡
            │ foo  ┆ Foo         │
            │ bar  ┆ Bar         │
            └──────┴─────────────┘

    On the testing front, there are a number of methods to help with testing the stage example. At the highest
    level of abstraction, you can call the `test` method to construct a test environment, run the stage
    through the MEDS-Transform CLI, and assert that the outputs are as expected. This, combined with the
    `stage_example` pytest fixtures in `../pytest_plugin.py`, gives an easy, accessible API to test stages
    that are registered with examples. If you need more flexibility in your usage, please file a GitHub Issue.

    To see the test code in action, though, we'll look at some lower layers of abstraction. First, we can set
    up and inspect the test environment with the `test_env` `contextmanager` property. This yields a `TestEnv`
    object, which has a nice readable `__str__` method, so we can print it to inspect it. Note that, upon
    creation, the `output_dir` part of the test environment is non-existent, but input data and config yaml
    files (if `do_use_config_yaml` is set to `True`) are created. The test environment is cleaned up when the
    context manager exits.

        >>> example = StageExample(stage_name="example_stage", want_metadata=metadata_df)
        >>> with example.test_env as test_env:
        ...     print(test_env)
        Test Environment in /tmp/tmp...
          - Files:
            └── input
                ├── data
                │   ├── held_out
                │   │   └── 0.parquet
                │   ├── train
                │   │   ├── 0.parquet
                │   │   └── 1.parquet
                │   └── tuning
                │       └── 0.parquet
                └── metadata
                    ├── codes.parquet
                    ├── dataset.json
                    └── subject_splits.parquet
          - Input sub-directory: input
          - Output sub-directory: cohort
          - Script: MEDS_transform-stage __null__ example_stage
                    input_dir=/tmp/tmp.../input output_dir=/tmp/tmp.../cohort
        >>> test_env.test_dir.is_dir()
        False
        >>> example = StageExample(
        ...     stage_name="example_stage",
        ...     want_metadata=metadata_df,
        ...     stage_cfg={"arg1": "value1", "arg2": {"arg3": ["value3A", "value3B"]}},
        ...     in_data=data,
        ...     do_use_config_yaml=True,
        ... )
        >>> with example.test_env as test_env:
        ...     print(test_env)
        Test Environment in /tmp/tmp...
          - Files:
            ├── config.yaml
            └── input
                ├── data
                │   └── 0.parquet
                └── metadata
                    ├── codes.parquet
                    └── dataset.json
          - Input sub-directory: input
          - Output sub-directory: cohort
          - Config yaml file: config.yaml
            │   stages:
            │   - example_stage:
            │       arg1: value1
            │       arg2:
            │         arg3:
            │         - value3A
            │         - value3B
          - Script: MEDS_transform-stage /tmp/tmp.../config.yaml example_stage
                    input_dir=/tmp/tmp.../input output_dir=/tmp/tmp.../cohort

    This test environment is used when we call the `test` method, which runs the stage, then checks to ensure
    that the stage ran successfully and the outputs in the `output_dir` are as expected. If the test fails, an
    error message including the string representation of the test environment and the command output is
    printed. To explore this, we'll make a fake run function that just returns a controllable return code and
    output.

        >>> def fake_run_success(script, shell, capture_output):
        ...     return subprocess.CompletedProcess([], returncode=0, stdout=b"Success", stderr=b"")
        >>> def fake_run_failure(script, shell, capture_output):
        ...     return subprocess.CompletedProcess([], returncode=1, stdout=b"", stderr=b"Failure")

    If we run and return a failing status code, the example will error as it is not expecting the command to
    fail:

        >>> example = StageExample(stage_name="example_stage", want_metadata=metadata_df)
        >>> example.test(run_fn=fake_run_failure)
        Traceback (most recent call last):
            ...
        AssertionError: Stage example example_stage Failed:
        Test Environment in /tmp/tmp...
          - Files:
            └── input
                ├── data
                │   ├── held_out
                │   │   └── 0.parquet
                │   ├── train
                │   │   ├── 0.parquet
                │   │   └── 1.parquet
                │   └── tuning
                │       └── 0.parquet
                └── metadata
                    ├── codes.parquet
                    ├── dataset.json
                    └── subject_splits.parquet
          - Input sub-directory: input
          - Output sub-directory: cohort
          - Script: MEDS_transform-stage __null__ example_stage
                    input_dir=/tmp/tmp.../input output_dir=/tmp/tmp.../cohort
        Stdout:
        <BLANKLINE>
        Stderr:
        Failure
        Command errored with return code 1

    If we have it succeed, it will still fail, as it won't find any of the output files it is expecting in the
    (non-existent) cohort directory:

        >>> example = StageExample(stage_name="example_stage", want_metadata=metadata_df)
        >>> example.test(run_fn=fake_run_success)
        Traceback (most recent call last):
            ...
        AssertionError: Stage example example_stage Failed:
        Test Environment in /tmp/tmp...
          - Files:
            └── input
                ├── data
                │   ├── held_out
                │   │   └── 0.parquet
                │   ├── train
                │   │   ├── 0.parquet
                │   │   └── 1.parquet
                │   └── tuning
                │       └── 0.parquet
                └── metadata
                    ├── codes.parquet
                    ├── dataset.json
                    └── subject_splits.parquet
          - Input sub-directory: input
          - Output sub-directory: cohort
          - Script: MEDS_transform-stage __null__ example_stage
                    input_dir=/tmp/tmp.../input output_dir=/tmp/tmp.../cohort
        Stdout:
        Success
        Stderr:
        <BLANKLINE>

    To explore test failures in more detail, we can use the `check_outputs` helper, which checks the output of
    the passed cohort directory is as expected. For example, if we add the expected metadata to a directory
    and validate that, we won't see any errors:

        >>> example = StageExample(stage_name="example_stage", want_metadata=metadata_df)
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     output_dir = Path(tmpdir)
        ...     (output_dir / "metadata").mkdir()
        ...     example.want_metadata.write_parquet(output_dir / code_metadata_filepath)
        ...     example.check_outputs(output_dir)
        ...     print("No error was raised!")
        No error was raised!

    This function will error if we don't have the expected metadata file in the cohort directory or if its
    contents are wrong. Note this function also returns a more minimal error, though the broader `test` helper
    wraps it in additional context.

        >>> example = StageExample(stage_name="example_stage", want_metadata=metadata_df)
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     output_dir = Path(tmpdir)
        ...     (output_dir / "metadata").mkdir()
        ...     example.check_outputs(output_dir)
        Traceback (most recent call last):
            ...
        AssertionError: Expected metadata file metadata/codes.parquet in /tmp/tmp.... Got:
        tmp...
        └── metadata
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     output_dir = Path(tmpdir)
        ...     (output_dir / "metadata").mkdir()
        ...     wrong_metadata = pl.DataFrame({"code": ["f"], "description": [None]})
        ...     wrong_metadata.write_parquet(output_dir / code_metadata_filepath)
        ...     example.check_outputs(output_dir)
        Traceback (most recent call last):
            ...
        AssertionError: Want metadata:
        shape: (2, 2)
        ┌──────┬─────────────┐
        │ code ┆ description │
        │ ---  ┆ ---         │
        │ str  ┆ str         │
        ╞══════╪═════════════╡
        │ foo  ┆ Foo         │
        │ bar  ┆ Bar         │
        └──────┴─────────────┘
        Got metadata:
        shape: (1, 2)
        ┌──────┬─────────────┐
        │ code ┆ description │
        │ ---  ┆ ---         │
        │ str  ┆ null        │
        ╞══════╪═════════════╡
        │ f    ┆ null        │
        └──────┴─────────────┘

    Similar assertion cases are used for data comparisons

        >>> data_df = pl.DataFrame(
        ...     {"subject_id": [1], "time": [datetime(2012, 12, 1)], "code": ["A"], "numeric_value": [None]},
        ...     schema_overrides={"numeric_value": pl.Float32},
        ... )
        >>> data = MEDSDataset(data_shards={"0": data_df}, dataset_metadata=DatasetMetadataSchema())
        >>> example = StageExample(stage_name="with_scenario", want_data=data)
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     output_dir = Path(tmpdir)
        ...     _ = example.want_data.write(output_dir)
        ...     example.check_outputs(output_dir)
        ...     print("No error was raised!")
        No error was raised!
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     output_dir = Path(tmpdir)
        ...     (output_dir / "data").mkdir()
        ...     (output_dir / "data" / "foo.json").write_text('{"foo": "bar"}')
        ...     example.check_outputs(output_dir)
        Traceback (most recent call last):
            ...
        AssertionError: Expected data files in /tmp/tmp.../data/**.parquet, but none were found. Got:
            tmp...
            └── data
                └── foo.json

    We also have an option to tell the system that we've already resolved the cohort dir into the appropriate
    data or metadata subdirectories; this is useful in pipeline tests. To do that, the `is_resolved_dir`
    parameter can be passed to `check_outputs` (it isn't passable to `test` as that is only used for
    single-stage tests which don't need this capability).

        >>> example = StageExample(stage_name="with_scenario", want_data=data)
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     output_dir = Path(tmpdir)
        ...     _ = example.want_data.write(output_dir)
        ...     example.check_outputs(output_dir / "data", is_resolved_dir=True)
        ...     print("No error was raised!")
        No error was raised!
        >>> example = StageExample(stage_name="example_stage", want_metadata=metadata_df)
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     output_dir = Path(tmpdir)
        ...     (output_dir / "metadata").mkdir()
        ...     example.want_metadata.write_parquet(output_dir / code_metadata_filepath)
        ...     example.check_outputs(output_dir / "metadata", is_resolved_dir=True)
        ...     print("No error was raised!")
        No error was raised!

    Different errors are raised if the shards differ...

        >>> example = StageExample(stage_name="with_scenario", want_data=data)
        >>> wrong_data = MEDSDataset(data_shards={"1": data_df}, dataset_metadata=DatasetMetadataSchema())
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     output_dir = Path(tmpdir)
        ...     _ = wrong_data.write(output_dir)
        ...     example.check_outputs(output_dir)
        Traceback (most recent call last):
            ...
        AssertionError: Want data:
            MEDSDataset:
            dataset_metadata:
            data_shards:
              - 0:
                pyarrow.Table
                subject_id: int64
                time: timestamp[us]
                code: string
                numeric_value: float
                ----
                subject_id: [[1]]
                time: [[2012-12-01 00:00:00.000000]]
                code: [["A"]]
                numeric_value: [[null]]
            code_metadata:
              pyarrow.Table
              code: string
              description: string
              parent_codes: list<item: string>
                child 0, item: string
              ----
              code: []
              description: []
              parent_codes: []
            subject_splits: None
            Got data:
            MEDSDataset:
            dataset_metadata:
            data_shards:
              - 1:
                pyarrow.Table
                subject_id: int64
                time: timestamp[us]
                code: string
                numeric_value: float
                ----
                subject_id: [[1]]
                time: [[2012-12-01 00:00:00.000000]]
                code: [["A"]]
                numeric_value: [[null]]
            code_metadata:
              pyarrow.Table
              code: string
              description: string
              parent_codes: list<item: string>
                child 0, item: string
              ----
              code: []
              description: []
              parent_codes: []
            subject_splits: None

    ...or if the contents of the data are different:

        >>> wrong_data_df = pl.DataFrame(
        ...     {"subject_id": [1], "time": [datetime(2015, 12, 1)], "code": ["A"], "numeric_value": [None]},
        ... )
        >>> wrong_data = MEDSDataset(
        ...     data_shards={"0": wrong_data_df}, dataset_metadata=DatasetMetadataSchema()
        ... )
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     output_dir = Path(tmpdir)
        ...     _ = wrong_data.write(output_dir)
        ...     example.check_outputs(output_dir)
        Traceback (most recent call last):
            ...
        AssertionError: Want data:
        MEDSDataset:
        dataset_metadata:
        data_shards:
          - 0:
            pyarrow.Table
            subject_id: int64
            time: timestamp[us]
            code: string
            numeric_value: float
            ----
            subject_id: [[1]]
            time: [[2012-12-01 00:00:00.000000]]
            code: [["A"]]
            numeric_value: [[null]]
        code_metadata:
          pyarrow.Table
          code: string
          description: string
          parent_codes: list<item: string>
            child 0, item: string
          ----
          code: []
          description: []
          parent_codes: []
        subject_splits: None
        Got data:
        MEDSDataset:
        dataset_metadata:
        data_shards:
          - 0:
            pyarrow.Table
            subject_id: int64
            time: timestamp[us]
            code: string
            numeric_value: float
            ----
            subject_id: [[1]]
            time: [[2015-12-01 00:00:00.000000]]
            code: [["A"]]
            numeric_value: [[null]]
        code_metadata:
          pyarrow.Table
          code: string
          description: string
          parent_codes: list<item: string>
            child 0, item: string
          ----
          code: []
          description: []
          parent_codes: []
        subject_splits: None
    """

    stage_name: str
    scenario_name: str | None = None
    stage_cfg: dict = field(default_factory=dict)
    want_data: MEDSDataset | None = None
    want_metadata: pl.DataFrame | None = None
    in_data: MEDSDataset | None = None
    do_use_config_yaml: bool = False
    df_check_kwargs: dict | None = None

    BASE_PACKAGE: ClassVar[str] = "__null__"

    def __post_init__(self):
        if self.want_data is None and self.want_metadata is None:
            raise ValueError("Either want_data or want_metadata must be provided.")
        elif self.want_data is not None and self.want_metadata is not None:
            raise ValueError("Either want_data or want_metadata must be provided, but not both.")

        if self.scenario_name == ".":
            self.scenario_name = None

        if self.df_check_kwargs is None:
            self.df_check_kwargs = {"rtol": 1e-3, "atol": 1e-5}

    @classmethod
    def is_example_dir(cls, path: Path) -> bool:
        """Check if the given path is a valid example directory."""
        want_data_fp = path / "out_data.yaml"
        want_metadata_fp = path / "out_metadata.yaml"
        return want_data_fp.is_file() or want_metadata_fp.is_file()

    @classmethod
    def from_dir(
        cls, stage_name: str, scenario_name: str, example_dir: Path, **schema_updates
    ) -> StageExample:
        """Parse the example directory and return a StageExample object, or raise an error if invalid."""

        stage_cfg_fp = example_dir / "cfg.yaml"
        in_fp = example_dir / "in.yaml"
        want_data_fp = example_dir / "out_data.yaml"
        want_metadata_fp = example_dir / "out_metadata.yaml"
        test_cfg_fp = example_dir / "_test_cfg.yaml"

        want_data = None
        want_metadata = None
        if want_data_fp.is_file():
            want_data = MEDSDataset.from_yaml(want_data_fp, **schema_updates)
        if want_metadata_fp.is_file():
            want_metadata = read_metadata_only(want_metadata_fp, **schema_updates)

        in_data = MEDSDataset.from_yaml(in_fp) if in_fp.is_file() else None
        stage_cfg = OmegaConf.to_container(OmegaConf.load(stage_cfg_fp)) if stage_cfg_fp.is_file() else {}
        test_kwargs = OmegaConf.to_container(OmegaConf.load(test_cfg_fp)) if test_cfg_fp.is_file() else {}

        return cls(
            stage_name=stage_name,
            scenario_name=scenario_name,
            want_data=want_data,
            want_metadata=want_metadata,
            stage_cfg=stage_cfg,
            in_data=in_data,
            **test_kwargs,
        )

    @property
    def full_name(self) -> str:
        if self.scenario_name:
            return f"{self.stage_name}/{self.scenario_name}"
        return self.stage_name

    def __str__(self) -> str:
        lines = [
            f"StageExample [{self.full_name}]",
            f"  stage_cfg: {self.stage_cfg}",
        ]

        if self.do_use_config_yaml:
            lines.append(f"  do_use_config_yaml: {self.do_use_config_yaml}")

        def add_nested_line(k: str):
            val = getattr(self, k)
            if val is not None:
                lines.append(f"  {k}:")
                lines.append(textwrap.indent(str(val), _SPACE))

        add_nested_line("in_data")
        add_nested_line("want_data")
        add_nested_line("want_metadata")

        return "\n".join(lines)

    def write_for_test(self, input_dir: Path) -> None:
        if self.in_data is None:
            in_data = MEDSDataset.from_yaml(SIMPLE_STATIC_SHARDED_BY_SPLIT)
            in_data.write(input_dir)
            return
        else:
            self.in_data.write(input_dir)
            self.in_data._pl_code_metadata.write_parquet(input_dir / code_metadata_filepath)
            for k, v in self.in_data._pl_shards.items():
                fp = input_dir / "data" / f"{k}.parquet"
                fp.parent.mkdir(parents=True, exist_ok=True)
                v.write_parquet(fp)

    def __data_files(self, data_dir: Path) -> list[Path]:
        return list((data_dir).rglob("*.parquet"))

    def __data_shards(self, data_dir: Path) -> dict[str, pl.DataFrame]:
        shards = {}
        for fp in self.__data_files(data_dir):
            shard_name = fp.relative_to(data_dir).with_suffix("").as_posix()
            shards[shard_name] = pl.read_parquet(fp)
        return shards

    def __check_files(self, output_dir: Path, is_resolved_dir: bool = False) -> None:
        if not output_dir.exists():
            raise AssertionError(f"Expected cohort directory {output_dir} to exist, but it does not.")

        all_files_str = f"{output_dir.name}\n" + "\n".join(list_directory(output_dir))

        if self.want_data is not None:
            data_dir = output_dir if is_resolved_dir else output_dir / "data"
            if not self.__data_files(data_dir):
                raise AssertionError(
                    f"Expected data files in {output_dir}/data/**.parquet, but none were found. Got:\n"
                    f"{all_files_str}"
                )

        if self.want_metadata is not None:
            metadata_dir = output_dir if is_resolved_dir else output_dir / "metadata"
            metadata_fp = metadata_dir / "codes.parquet"
            if not metadata_fp.is_file():
                raise AssertionError(
                    f"Expected metadata file {code_metadata_filepath} in {output_dir}. Got:\n{all_files_str}"
                )

    def check_outputs(self, output_dir: Path, is_resolved_dir: bool = False) -> None:
        self.__check_files(output_dir, is_resolved_dir)

        if self.want_data is not None:
            data_dir = output_dir if is_resolved_dir else output_dir / "data"
            got_data = MEDSDataset(
                data_shards=self.__data_shards(data_dir), dataset_metadata=DatasetMetadataSchema()
            )

            try:
                assert got_data._pl_shards.keys() == self.want_data._pl_shards.keys(), (
                    f"Shards differ: {got_data._pl_shards.keys()} vs {self.want_data._pl_shards.keys()}"
                )
                for shard_name, got_df in got_data._pl_shards.items():
                    want_df = self.want_data._pl_shards[shard_name]
                    try:
                        assert_frame_equal(got_df, want_df, **self.df_check_kwargs)
                    except AssertionError as e:
                        raise AssertionError(f"Data differs in (at least) shard {shard_name}") from e
            except AssertionError as e:
                pl.Config.set_tbl_rows(-1)
                raise AssertionError(f"Want data:\n{self.want_data}\nGot data:\n{got_data}") from e

        if self.want_metadata is not None:
            metadata_dir = output_dir if is_resolved_dir else output_dir / "metadata"
            metadata_fp = metadata_dir / "codes.parquet"
            got_metadata = pl.read_parquet(metadata_fp)
            try:
                assert_frame_equal(self.want_metadata, got_metadata, **self.df_check_kwargs)
            except AssertionError as e:
                pl.Config.set_tbl_rows(-1)
                raise AssertionError(
                    f"Want metadata:\n{self.want_metadata}\nGot metadata:\n{got_metadata}"
                ) from e

    @property
    def cmd_pipeline_cfg(self) -> DictConfig | None:
        if not self.do_use_config_yaml:
            return None

        return OmegaConf.create({"stages": [{self.stage_name: self.stage_cfg}]})

    @property
    def cmd_args(self) -> list[str]:
        return [] if self.do_use_config_yaml else dict_to_hydra_kwargs({"stage_cfg": self.stage_cfg})

    @property
    @contextmanager
    def test_env(self) -> TestEnv:
        """Context manager to set up the test environment.

        This sets up the test directory and cleans it up after use. It yields the script to run this stage and
        the cohort directory to check the outputs.
        """
        with tempfile.TemporaryDirectory() as test_dir:
            test_dir = Path(test_dir)

            input_dir = test_dir / "input"
            input_dir.mkdir()

            self.write_for_test(input_dir)

            output_dir = test_dir / "cohort"

            if self.do_use_config_yaml:
                cfg_yaml_fp = test_dir / "config.yaml"
                OmegaConf.save(self.cmd_pipeline_cfg, cfg_yaml_fp)
                pipeline_cfg_yaml = cfg_yaml_fp.resolve()
            else:
                pipeline_cfg_yaml = self.BASE_PACKAGE

            script = (
                f"MEDS_transform-stage {pipeline_cfg_yaml} {self.stage_name} "
                f"{' '.join(self.cmd_args)} input_dir={input_dir} output_dir={output_dir}"
            )

            yield TestEnv(
                script=script,
                output_dir=output_dir,
                test_dir=test_dir,
                input_dir=input_dir,
                config_yaml_fp=pipeline_cfg_yaml if self.do_use_config_yaml else None,
            )

    @property
    def _err_prefix(self) -> str:
        lines = [f"Stage example {self.full_name} Failed:"]
        if self.do_use_config_yaml:
            lines.append(f"Config:\n{self.cmd_pipeline_cfg}")
        return "\n".join(lines)

    def test(self, run_fn: Callable | None = subprocess.run) -> None:
        """Run a test for this example and assert correctness.

        Args:
            run_fn: The function to use to run the command. Defaults to subprocess.run. This is useful for
                dependency injection or customizing the test set-up.

        Raises:
            AssertionError: If the test fails.
        """

        with self.test_env as test_env:
            command_out = run_fn(test_env.script, shell=True, capture_output=True)
            err_lines = [self._err_prefix, *str(test_env).splitlines()]

            err_lines.append(f"Stdout:\n{command_out.stdout.decode()}")
            err_lines.append(f"Stderr:\n{command_out.stderr.decode()}")

            if command_out.returncode != 0:
                err_lines.append(f"Command errored with return code {command_out.returncode}")
                raise AssertionError("\n".join(err_lines))

            try:
                self.check_outputs(test_env.output_dir)
            except AssertionError as e:
                raise AssertionError("\n".join(err_lines)) from e


class StageExampleDict(dict):
    """A dictionary subclass to hold stage examples.

    The only purpose to this is to display the examples in a more readable format when printed. It does not
    validate that the values are `StageExample` objects,

    Examples:
        >>> print(StageExampleDict())
        {}

    Singleton example dictionaries with an empty scenario name just prints the stage:

        >>> print(StageExampleDict(**{".": "foo"}))
        foo

    With a scenario name or multiple scenarios, it prints the stage and scenario, nesting with a "│" indicator

        >>> print(StageExampleDict(**{"foo": "bar"}))
        foo:
        │   bar

    Note that if the empty scenario name (".") is included in a multi-scenario dict, it is printed, unlike in
    the singleton case.

        >>> print(StageExampleDict(**{"foo": "bar", "baz": "qux", ".": "quux"}))
        foo:
        │   bar
        baz:
        │   qux
        .:
        │   quux
    """

    def __str__(self) -> str:
        if not self:
            return "{}"

        if len(self) == 1 and next(iter(self.keys())) == ".":
            return str(next(iter(self.values())))

        lines = []
        for k, v in self.items():
            lines.append(f"{k}:")
            v_str = textwrap.indent(str(v), _BRANCH)

            lines.append(v_str)

        return "\n".join(lines)



"""A functor to compute time-of-day measurements on a MEDS dataset."""

import logging
from collections.abc import Callable

import polars as pl
from meds import DataSchema
from omegaconf import DictConfig

from .utils import unique_events

logger = logging.getLogger(__name__)


def time_of_day_fntr(cfg: DictConfig) -> Callable[[pl.DataFrame], pl.DataFrame]:
    """Create a function that adds the time of day to a DataFrame.

    Args:
        cfg: The configuration for the time of day function. This must contain an "endpoints" key with the
            endpoints (in hours of the day in 24-hour time) for each time of day category and a
            "time_of_day_code" key for the code for the time of day event in the output data.

    Returns:
        A function that adds the time of day to a DataFrame.

    Examples:
        >>> from datetime import datetime
        >>> df = pl.DataFrame(
        ...     {
        ...         "subject_id": [1, 1, 1, 1, 2, 2, 3, 3],
        ...         "time": [
        ...             None,
        ...             datetime(1990, 1, 1, 1, 0),
        ...             datetime(2021, 1, 1, 12, 0),
        ...             datetime(2021, 1, 2, 23, 59),
        ...             datetime(1988, 1, 2, 6, 0),
        ...             datetime(2023, 1, 3, 12, 0),
        ...             datetime(2022, 1, 1, 18, 0),
        ...             datetime(2022, 1, 1, 18, 0),
        ...         ],
        ...         "code": ["static", "DOB", "lab//A", "lab//B", "DOB", "lab//A", "lab//B", "dx//1"],
        ...     },
        ...     schema={"subject_id": pl.UInt32, "time": pl.Datetime, "code": pl.Utf8},
        ... )
        >>> df
        shape: (8, 3)
        ┌────────────┬─────────────────────┬────────┐
        │ subject_id ┆ time                ┆ code   │
        │ ---        ┆ ---                 ┆ ---    │
        │ u32        ┆ datetime[μs]        ┆ str    │
        ╞════════════╪═════════════════════╪════════╡
        │ 1          ┆ null                ┆ static │
        │ 1          ┆ 1990-01-01 01:00:00 ┆ DOB    │
        │ 1          ┆ 2021-01-01 12:00:00 ┆ lab//A │
        │ 1          ┆ 2021-01-02 23:59:00 ┆ lab//B │
        │ 2          ┆ 1988-01-02 06:00:00 ┆ DOB    │
        │ 2          ┆ 2023-01-03 12:00:00 ┆ lab//A │
        │ 3          ┆ 2022-01-01 18:00:00 ┆ lab//B │
        │ 3          ┆ 2022-01-01 18:00:00 ┆ dx//1  │
        └────────────┴─────────────────────┴────────┘
        >>> time_of_day_cfg = DictConfig({"time_of_day_code": "time_of_day", "endpoints": [6, 12, 18]})
        >>> time_of_day_fn = time_of_day_fntr(time_of_day_cfg)
        >>> time_of_day_fn(df)
        shape: (6, 3)
        ┌────────────┬─────────────────────┬──────────────────────┐
        │ subject_id ┆ time                ┆ code                 │
        │ ---        ┆ ---                 ┆ ---                  │
        │ u32        ┆ datetime[μs]        ┆ str                  │
        ╞════════════╪═════════════════════╪══════════════════════╡
        │ 1          ┆ 1990-01-01 01:00:00 ┆ time_of_day//[00,06) │
        │ 1          ┆ 2021-01-01 12:00:00 ┆ time_of_day//[12,18) │
        │ 1          ┆ 2021-01-02 23:59:00 ┆ time_of_day//[18,24) │
        │ 2          ┆ 1988-01-02 06:00:00 ┆ time_of_day//[06,12) │
        │ 2          ┆ 2023-01-03 12:00:00 ┆ time_of_day//[12,18) │
        │ 3          ┆ 2022-01-01 18:00:00 ┆ time_of_day//[18,24) │
        └────────────┴─────────────────────┴──────────────────────┘
        >>> time_of_day_fntr(DictConfig({"endpoints": []}))
        Traceback (most recent call last):
            ...
        ValueError: The 'endpoints' key must contain at least one endpoint for time of day categories.
        >>> time_of_day_fntr(DictConfig({"endpoints": [6, 12, 36]}))
        Traceback (most recent call last):
            ...
        ValueError: All endpoints must be between 0 and 24 inclusive. Got: [6, 12, 36]
        >>> time_of_day_fntr(DictConfig({"endpoints": [6, 1.2]}))
        Traceback (most recent call last):
            ...
        ValueError: All endpoints must be integer, whole-hour boundaries, but got: [6, 1.2]
        >>> time_of_day_fntr(DictConfig({"endpoints": [6, 6]}))
        Traceback (most recent call last):
            ...
        ValueError: All endpoints must be unique. Got: [6, 6]
        >>> time_of_day_fntr(DictConfig({"endpoints": [6, 12, 10]}))
        Traceback (most recent call last):
            ...
        ValueError: All endpoints must be in sorted order. Got: [6, 12, 10]
    """
    if not cfg.endpoints:
        raise ValueError("The 'endpoints' key must contain at least one endpoint for time of day categories.")
    if not all(0 <= endpoint <= 24 for endpoint in cfg.endpoints):
        raise ValueError(f"All endpoints must be between 0 and 24 inclusive. Got: {cfg.endpoints}")
    if not all(isinstance(endpoint, int) for endpoint in cfg.endpoints):
        raise ValueError(f"All endpoints must be integer, whole-hour boundaries, but got: {cfg.endpoints}")
    if len(cfg.endpoints) != len(set(cfg.endpoints)):
        raise ValueError(f"All endpoints must be unique. Got: {cfg.endpoints}")
    if cfg.endpoints != sorted(cfg.endpoints):
        raise ValueError(f"All endpoints must be in sorted order. Got: {cfg.endpoints}")

    def fn(df: pl.LazyFrame) -> pl.LazyFrame:
        hour = pl.col(DataSchema.time_name).dt.hour()
        code_dtype = df.collect_schema().get(DataSchema.code_name, pl.Utf8)

        def tod_code(start: int, end: int) -> str:
            return pl.lit(f"{cfg.time_of_day_code}//[{start:02},{end:02})", dtype=code_dtype)

        start, end = 0, cfg.endpoints[0]
        time_of_day = pl.when(hour < end).then(tod_code(start, end))

        for i in range(1, len(cfg.endpoints)):
            start, end = cfg.endpoints[i - 1], cfg.endpoints[i]

            time_of_day = time_of_day.when((hour >= start) & (hour < end)).then(tod_code(start, end))

        time_of_day = time_of_day.when(hour >= end).then(tod_code(end, 24))
        return unique_events(df).with_columns(time_of_day.alias(DataSchema.code_name))

    return fn



"""A polars-to-polars transformation function for filtering subjects by sequence length."""

import logging
from collections.abc import Callable
from functools import partial

import polars as pl
from omegaconf import DictConfig

from .. import Stage

logger = logging.getLogger(__name__)


def filter_subjects_by_num_measurements(df: pl.LazyFrame, min_measurements_per_subject: int) -> pl.LazyFrame:
    """Filters subjects by the number of dynamic (timestamp non-null) measurements they have.

    Args:
        df: The input DataFrame.
        min_measurements_per_subject: The minimum number of measurements a subject must have to be included.

    Returns:
        The filtered DataFrame.

    Examples:
        >>> df = pl.DataFrame(
        ...     {
        ...         "subject_id": [1, 1, 1, 2, 2, 3, 3, 4],
        ...         "time": [1, 2, 1, 1, 2, 1, None, None],
        ...     }
        ... )
        >>> filter_subjects_by_num_measurements(df, 1)
        shape: (7, 2)
        ┌────────────┬──────┐
        │ subject_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 1          ┆ 1    │
        │ 1          ┆ 2    │
        │ 1          ┆ 1    │
        │ 2          ┆ 1    │
        │ 2          ┆ 2    │
        │ 3          ┆ 1    │
        │ 3          ┆ null │
        └────────────┴──────┘
        >>> filter_subjects_by_num_measurements(df, 2)
        shape: (5, 2)
        ┌────────────┬──────┐
        │ subject_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 1          ┆ 1    │
        │ 1          ┆ 2    │
        │ 1          ┆ 1    │
        │ 2          ┆ 1    │
        │ 2          ┆ 2    │
        └────────────┴──────┘
        >>> filter_subjects_by_num_measurements(df, 3)
        shape: (3, 2)
        ┌────────────┬──────┐
        │ subject_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 1          ┆ 1    │
        │ 1          ┆ 2    │
        │ 1          ┆ 1    │
        └────────────┴──────┘
        >>> filter_subjects_by_num_measurements(df, 4)
        shape: (0, 2)
        ┌────────────┬──────┐
        │ subject_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        └────────────┴──────┘
        >>> filter_subjects_by_num_measurements(df, 2.2)
        Traceback (most recent call last):
            ...
        TypeError: min_measurements_per_subject must be an integer; got <class 'float'> 2.2
    """
    if not isinstance(min_measurements_per_subject, int):
        raise TypeError(
            f"min_measurements_per_subject must be an integer; got {type(min_measurements_per_subject)} "
            f"{min_measurements_per_subject}"
        )

    return df.filter(pl.col("time").count().over("subject_id") >= min_measurements_per_subject)


def filter_subjects_by_num_events(df: pl.LazyFrame, min_events_per_subject: int) -> pl.LazyFrame:
    """Filters subjects by the number of events (unique timepoints) they have.

    Args:
        df: The input DataFrame.
        min_events_per_subject: The minimum number of events a subject must have to be included.

    Returns:
        The filtered DataFrame.

    Examples:
        >>> df = pl.DataFrame(
        ...     {
        ...         "subject_id": [1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 4],
        ...         "time": [1, 1, 1, 1, 2, 1, 1, 2, 3, None, None, 1, 2, 3],
        ...     }
        ... )
        >>> with pl.Config(tbl_rows=15):
        ...     filter_subjects_by_num_events(df, 1)
        shape: (14, 2)
        ┌────────────┬──────┐
        │ subject_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 1          ┆ 1    │
        │ 1          ┆ 1    │
        │ 1          ┆ 1    │
        │ 2          ┆ 1    │
        │ 2          ┆ 2    │
        │ 2          ┆ 1    │
        │ 3          ┆ 1    │
        │ 3          ┆ 2    │
        │ 3          ┆ 3    │
        │ 4          ┆ null │
        │ 4          ┆ null │
        │ 4          ┆ 1    │
        │ 4          ┆ 2    │
        │ 4          ┆ 3    │
        └────────────┴──────┘
        >>> with pl.Config(tbl_rows=15):
        ...     filter_subjects_by_num_events(df, 2)
        shape: (11, 2)
        ┌────────────┬──────┐
        │ subject_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 2          ┆ 1    │
        │ 2          ┆ 2    │
        │ 2          ┆ 1    │
        │ 3          ┆ 1    │
        │ 3          ┆ 2    │
        │ 3          ┆ 3    │
        │ 4          ┆ null │
        │ 4          ┆ null │
        │ 4          ┆ 1    │
        │ 4          ┆ 2    │
        │ 4          ┆ 3    │
        └────────────┴──────┘
        >>> filter_subjects_by_num_events(df, 3)
        shape: (8, 2)
        ┌────────────┬──────┐
        │ subject_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 3          ┆ 1    │
        │ 3          ┆ 2    │
        │ 3          ┆ 3    │
        │ 4          ┆ null │
        │ 4          ┆ null │
        │ 4          ┆ 1    │
        │ 4          ┆ 2    │
        │ 4          ┆ 3    │
        └────────────┴──────┘
        >>> filter_subjects_by_num_events(df, 4)
        shape: (5, 2)
        ┌────────────┬──────┐
        │ subject_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 4          ┆ null │
        │ 4          ┆ null │
        │ 4          ┆ 1    │
        │ 4          ┆ 2    │
        │ 4          ┆ 3    │
        └────────────┴──────┘
        >>> filter_subjects_by_num_events(df, 5)
        shape: (0, 2)
        ┌────────────┬──────┐
        │ subject_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        └────────────┴──────┘
        >>> filter_subjects_by_num_events(df, 2.2)
        Traceback (most recent call last):
            ...
        TypeError: min_events_per_subject must be an integer; got <class 'float'> 2.2
    """
    if not isinstance(min_events_per_subject, int):
        raise TypeError(
            f"min_events_per_subject must be an integer; got {type(min_events_per_subject)} "
            f"{min_events_per_subject}"
        )

    return df.filter(pl.col("time").n_unique().over("subject_id") >= min_events_per_subject)


@Stage.register
def filter_subjects(stage_cfg: DictConfig) -> Callable[[pl.LazyFrame], pl.LazyFrame]:
    """Returns a function that filters subjects by the number of measurements and events they have.

    Args:
        stage_cfg: The stage configuration. Arguments include: min_measurements_per_subject,
            min_events_per_subject, both of which should be integers or None which specify the minimum number
            of measurements and events a subject must have to be included, respectively.

    Returns:
        The function that filters subjects by the number of measurements and/or events they have.

    Examples:
        >>> df = pl.DataFrame(
        ...     {
        ...         "subject_id": [1, 1, 1, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5],
        ...         "time": [1, 1, 1, 1, 1, 1, 2, 3, None, None, 1, 2, 2, None, 1, 2, 3, 1],
        ...     }
        ... )
        >>> stage_cfg = DictConfig({"min_measurements_per_subject": 4, "min_events_per_subject": 2})
        >>> filter_subjects(stage_cfg)(df)
        shape: (4, 2)
        ┌────────────┬──────┐
        │ subject_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 5          ┆ 1    │
        │ 5          ┆ 2    │
        │ 5          ┆ 3    │
        │ 5          ┆ 1    │
        └────────────┴──────┘
    """
    map_fns = []
    if stage_cfg.min_measurements_per_subject:
        logger.info(
            f"Filtering subjects with fewer than {stage_cfg.min_measurements_per_subject} measurements "
            "(observations of any kind)."
        )
        map_fns.append(
            partial(
                filter_subjects_by_num_measurements,
                min_measurements_per_subject=stage_cfg.min_measurements_per_subject,
            )
        )
    if stage_cfg.min_events_per_subject:
        logger.info(
            f"Filtering subjects with fewer than {stage_cfg.min_events_per_subject} events "
            "(unique timepoints)."
        )
        map_fns.append(
            partial(filter_subjects_by_num_events, min_events_per_subject=stage_cfg.min_events_per_subject)
        )

    def fn(data: pl.LazyFrame) -> pl.LazyFrame:
        for map_fn in map_fns:
            data = map_fn(data)
        return data

    return fn



"""This file defines an object-oriented backing for configuring stages in MEDS-Transforms pipelines."""

from __future__ import annotations

import dataclasses
from typing import ClassVar

from omegaconf import DictConfig

UNPARSED_STAGE_T = str | dict | DictConfig


@dataclasses.dataclass
class StageConfig:
    """A configuration class for a stage in the MEDS-transforms pipeline.

    This class is used to define the structure of a stage configuration options within a `PipelineConfig`. It
    is generally constructed from raw pipeline arguments through the `from_arg` classmethod, which parses a
    raw entry into structured form. See that method for more detailed examples of the accepted format.

    Attributes:
        name: The name of the stage.
        base_stage: The name of the underlying, registered, base stage (if any) for this stage.
        config: The configuration for the stage.

    Raises:
        TypeError: If any non-None parameter is not of the appropriate type.
        ValueError: If the base stage name is empty.

    Examples:
        >>> StageConfig(name="stage_name")
        StageConfig(name='stage_name', base_stage=None, config={})
        >>> StageConfig(name="stage_name", base_stage="base_stage", config={"param1": 1})
        StageConfig(name='stage_name', base_stage='base_stage', config={'param1': 1})

    `StageConfig`s also expose a property `resolved_name`, which returns the resolved name of the stage (i.e.,
    the base name if it exists, otherwise the name itself):

        >>> cfg = StageConfig(name="foo", base_stage="bar")
        >>> cfg.resolved_name
        'bar'
        >>> cfg = StageConfig(name="foo", config={"param1": 1})
        >>> cfg.resolved_name
        'foo'

    Errors are raised if types are invalid or the base stage name is empty:

        >>> StageConfig(name=3)
        Traceback (most recent call last):
            ...
        TypeError: Invalid stage name type <class 'int'>. Expected str.
        >>> StageConfig(name="foo", base_stage=3)
        Traceback (most recent call last):
            ...
        TypeError: Invalid base stage type <class 'int'>. Expected str.
        >>> StageConfig(name="foo", base_stage="")
        Traceback (most recent call last):
            ...
        ValueError: If specified, base stage name cannot be empty.
        >>> StageConfig(name="foo", config=3)
        Traceback (most recent call last):
            ...
        TypeError: Invalid config type <class 'int'>. Expected dict or DictConfig.
    """

    name: str
    base_stage: str | None = None
    config: dict | DictConfig = dataclasses.field(default_factory=dict)

    META_KEYS: ClassVar[set[str]] = {
        "_base_stage",
    }

    @classmethod
    def _is_meta_key(cls, key: str) -> bool:
        """Checks if a key is a meta-key for the configuration.

        Meta-keys are keys that are in the `cls.META_KEYS` set.

        Args:
            key: The key to check.

        Returns:
            True if the key is a meta-key, False otherwise.

        Examples:
            >>> StageConfig._is_meta_key("_base_stage")
            True
            >>> StageConfig._is_meta_key("base_stage")
            False
            >>> StageConfig._is_meta_key("_foo")
            False
        """
        return key in cls.META_KEYS

    @classmethod
    def _split_meta_keys(cls, config: dict | DictConfig) -> tuple[dict, dict]:
        """Splits the meta-keys from the rest of the configuration.

        Args:
            config: The configuration to split.

        Returns:
            A tuple containing the dictionary's meta-keys and the rest of the configuration.

        Examples:
            >>> StageConfig._split_meta_keys({"_base_stage": "base_stage", "param1": 1})
            ({'_base_stage': 'base_stage'}, {'param1': 1})
            >>> StageConfig._split_meta_keys({"param1": 1})
            ({}, {'param1': 1})
            >>> StageConfig._split_meta_keys({"_base_stage": "base_stage"})
            ({'_base_stage': 'base_stage'}, {})
        """
        meta_keys = {}
        non_meta_keys = {}

        for k, v in config.items():
            if StageConfig._is_meta_key(k):
                meta_keys[k] = v
            else:
                non_meta_keys[k] = v

        return meta_keys, non_meta_keys

    @classmethod
    def _dict_arg_error_str(cls, arg: dict | DictConfig) -> str | None:
        """Checks if the argument is a valid dictionary representation of a stage configuration.

        Args:
            arg: The argument to check.

        Returns:
            True if the argument is a valid dictionary representation, False otherwise. A valid dictionary has
                all keys as strings and has only one key that is not a meta-key, which points to a dictionary.

        Examples:
            >>> print(StageConfig._dict_arg_error_str({"stage": {"param1": 1}}))
            None
            >>> print(StageConfig._dict_arg_error_str({"stage": {"_base_stage": "base_stage", "param1": 1}}))
            None
            >>> print(StageConfig._dict_arg_error_str({"stage": {"foo": "foobar"}, "_base_stage": "base"}))
            None
            >>> print(StageConfig._dict_arg_error_str({123: {"param1": 1}}))
            All keys must be strings. Got key(s):
              - int: 123
            >>> print(StageConfig._dict_arg_error_str({"stage": {"param1": 1}, "param2": 2}))
            Expected a single non-meta key-value pair representing the stage name and configuration. Got 2
            non-meta keys: stage, param2.
            >>> print(StageConfig._dict_arg_error_str({"S": {"_base_stage": "foo"}, "_base_stage": "foo"}))
            You cannot specify meta keys in both the raw and stage configuration dictionaries. Got duplicate
            meta keys: _base_stage.
        """
        if any(not isinstance(k, str) for k in arg):
            non_str_keys = [k for k in arg if not isinstance(k, str)]
            err_strs = [f"  - {type(key).__name__}: {key}" for key in non_str_keys]
            err_str = "\n".join(err_strs)
            return f"All keys must be strings. Got key(s):\n{err_str}"

        meta_keys, non_meta_keys = cls._split_meta_keys(arg)
        if len(non_meta_keys) != 1:
            return (
                "Expected a single non-meta key-value pair representing the stage name and configuration. "
                f"Got {len(non_meta_keys)} non-meta keys: {', '.join(non_meta_keys.keys())}."
            )

        stage_config = next(iter(non_meta_keys.values()))

        meta_key_duplicates = [k for k in meta_keys if k in stage_config]
        if meta_key_duplicates:
            return (
                "You cannot specify meta keys in both the raw and stage configuration dictionaries. "
                f"Got duplicate meta keys: {', '.join(meta_key_duplicates)}."
            )
        return None

    @classmethod
    def from_arg(cls, arg: UNPARSED_STAGE_T) -> StageConfig:
        """Resolves a `StageConfig` object from a given argument (yaml file) representation.

        Args:
            arg: The argument to resolve. This can be a string representing the stage name, indicating no
                additional configuration arguments, a dictionary representing the stage configuration in
                resolved form, or a dictionary with a single key-value, non-meta-key pair representing the
                stage name and its configuration, with meta-keys (e.g., `_base_stage`) being pulled from the
                base dict or the stage configuration as appropriate.

        Returns:
            A `StageConfig` object representing the resolved stage configuration.

        Raises:
            ValueError: If the argument is not a valid stage configuration.
            TypeError: If the argument is not a string, dictionary, or `DictConfig`.

        Examples:
            >>> StageConfig.from_arg("stage_name")
            StageConfig(name='stage_name', base_stage=None, config={})
            >>> StageConfig.from_arg({"stage1": {"_base_stage": "foo"}})
            StageConfig(name='stage1', base_stage='foo', config={})
            >>> StageConfig.from_arg({"stage_name": {"param1": 1}})
            StageConfig(name='stage_name', base_stage=None, config={'param1': 1})
            >>> StageConfig.from_arg({"stage_name": {"_base_stage": "base_stage", "param1": 1}})
            StageConfig(name='stage_name', base_stage='base_stage', config={'param1': 1})
            >>> StageConfig.from_arg({"foobar": {"param1": 1}, "_base_stage": "barfoo"})
            StageConfig(name='foobar', base_stage='barfoo', config={'param1': 1})

            Errors are thrown if the argument is the wrong type:

            >>> StageConfig.from_arg(3)
            Traceback (most recent call last):
                ...
            TypeError: Invalid argument type <class 'int'>. Expected str, dict, or DictConfig.

            Or if the argument is not a valid stage configuration. For example, if the argument is not a
            single (non-meta) key-value pair:

            >>> StageConfig.from_arg({"stage_name": {"param1": 1}, "param2": 2})
            Traceback (most recent call last):
                ...
            ValueError: Invalid stage config: Expected a single non-meta key-value pair representing the stage
            name and configuration. Got 2 non-meta keys: stage_name, param2.

            Or if the argument is not a valid stage name, or if meta-keys are used improperly:

            >>> StageConfig.from_arg({123: {"param1": 1}})
            Traceback (most recent call last):
                ...
            ValueError: Invalid stage config: All keys must be strings. Got key(s):
              - int: 123
            >>> StageConfig.from_arg({"stage_name": {"_base_stage": "foobar"}, "_base_stage": "barfoo"})
            Traceback (most recent call last):
                ...
            ValueError: Invalid stage config: You cannot specify meta keys in both the raw and stage
                configuration dictionaries. Got duplicate meta keys: _base_stage.
        """

        match arg:
            case str() as name:
                return cls(name=name)
            case dict() | DictConfig() as config:
                if error_str := cls._dict_arg_error_str(config):
                    raise ValueError(f"Invalid stage config: {error_str}")

                meta_keys, non_meta_keys = cls._split_meta_keys(config)
                stage_name = next(iter(non_meta_keys.keys()))
                stage_config = non_meta_keys[stage_name]

                meta_keys = {k[1:]: v for k, v in meta_keys.items()}
                for k in cls.META_KEYS:
                    if k in stage_config:
                        meta_keys[k[1:]] = stage_config.pop(k)

                return cls(name=stage_name, config=stage_config, **meta_keys)
            case _:
                raise TypeError(f"Invalid argument type {type(arg)}. Expected str, dict, or DictConfig.")

    def __post_init__(self):
        if not isinstance(self.name, str):
            raise TypeError(f"Invalid stage name type {type(self.name)}. Expected str.")
        if self.base_stage is not None:
            if not isinstance(self.base_stage, str):
                raise TypeError(f"Invalid base stage type {type(self.base_stage)}. Expected str.")
            if not self.base_stage:
                raise ValueError("If specified, base stage name cannot be empty.")

        if not isinstance(self.config, dict | DictConfig):
            raise TypeError(f"Invalid config type {type(self.config)}. Expected dict or DictConfig.")

    @property
    def resolved_name(self) -> str:
        """Return the resolved (base_name, if it exists, otherwise name) of the stage configuration."""
        return self.base_stage if self.base_stage is not None else self.name



"""Transformations for adding time-derived measurements (e.g., a subject's age) to a MEDS dataset."""

import logging
from collections.abc import Callable

import polars as pl
from meds import DataSchema
from omegaconf import DictConfig, OmegaConf

from ... import INFERRED_STAGE_KEYS
from .. import Stage
from .age import age_fntr
from .time_of_day import time_of_day_fntr
from .timeline_tokens import timeline_tokens_fntr

logger = logging.getLogger(__name__)


def add_new_events_fntr(
    fn: Callable[[pl.DataFrame], pl.DataFrame],
    new_code_last_regex: str | None = None,
) -> Callable[[pl.DataFrame], pl.DataFrame]:
    """Creates a "meta" functor that computes the input functor on a MEDS shard then combines both dataframes.

    Args:
        fn: The function that computes the new events.

    Returns:
        A function that computes the new events and combines them with the original DataFrame, returning a
        result in proper MEDS sorted order.

    Examples:
        >>> from datetime import datetime
        >>> df = pl.DataFrame(
        ...     {
        ...         "subject_id": [1, 1, 1, 1, 2, 2, 3, 3],
        ...         "time": [
        ...             None,
        ...             datetime(1990, 1, 1),
        ...             datetime(2021, 1, 1),
        ...             datetime(2021, 1, 1),
        ...             datetime(1988, 1, 2),
        ...             datetime(2023, 1, 3),
        ...             datetime(2022, 1, 1),
        ...             datetime(2022, 1, 1),
        ...         ],
        ...         "code": ["static", "DOB", "lab//A", "lab//B", "DOB", "lab//A", "lab//B", "dx//1"],
        ...     },
        ...     schema={"subject_id": pl.UInt32, "time": pl.Datetime, "code": pl.Utf8},
        ... )
        >>> df
        shape: (8, 3)
        ┌────────────┬─────────────────────┬────────┐
        │ subject_id ┆ time                ┆ code   │
        │ ---        ┆ ---                 ┆ ---    │
        │ u32        ┆ datetime[μs]        ┆ str    │
        ╞════════════╪═════════════════════╪════════╡
        │ 1          ┆ null                ┆ static │
        │ 1          ┆ 1990-01-01 00:00:00 ┆ DOB    │
        │ 1          ┆ 2021-01-01 00:00:00 ┆ lab//A │
        │ 1          ┆ 2021-01-01 00:00:00 ┆ lab//B │
        │ 2          ┆ 1988-01-02 00:00:00 ┆ DOB    │
        │ 2          ┆ 2023-01-03 00:00:00 ┆ lab//A │
        │ 3          ┆ 2022-01-01 00:00:00 ┆ lab//B │
        │ 3          ┆ 2022-01-01 00:00:00 ┆ dx//1  │
        └────────────┴─────────────────────┴────────┘

    As an example, we'll use the `age_fntr` defined in this module:

        >>> age_cfg = DictConfig({"DOB_code": "DOB", "age_code": "AGE", "age_unit": "years"})
        >>> age_fn = age_fntr(age_cfg)
        >>> age_fn(df)
        shape: (2, 4)
        ┌────────────┬─────────────────────┬──────┬───────────────┐
        │ subject_id ┆ time                ┆ code ┆ numeric_value │
        │ ---        ┆ ---                 ┆ ---  ┆ ---           │
        │ u32        ┆ datetime[μs]        ┆ str  ┆ f32           │
        ╞════════════╪═════════════════════╪══════╪═══════════════╡
        │ 1          ┆ 2021-01-01 00:00:00 ┆ AGE  ┆ 31.001347     │
        │ 2          ┆ 2023-01-03 00:00:00 ┆ AGE  ┆ 35.004169     │
        └────────────┴─────────────────────┴──────┴───────────────┘

    If we use the add_new_events_fntr on the age_fn, we'll get a function that computes and adds AGE events to
    the beginning of each event row block for the subjects.

        >>> add_age_fn = add_new_events_fntr(age_fn)
        >>> add_age_fn(df)
        shape: (10, 4)
        ┌────────────┬─────────────────────┬────────┬───────────────┐
        │ subject_id ┆ time                ┆ code   ┆ numeric_value │
        │ ---        ┆ ---                 ┆ ---    ┆ ---           │
        │ u32        ┆ datetime[μs]        ┆ str    ┆ f32           │
        ╞════════════╪═════════════════════╪════════╪═══════════════╡
        │ 1          ┆ null                ┆ static ┆ null          │
        │ 1          ┆ 1990-01-01 00:00:00 ┆ DOB    ┆ null          │
        │ 1          ┆ 2021-01-01 00:00:00 ┆ AGE    ┆ 31.001347     │
        │ 1          ┆ 2021-01-01 00:00:00 ┆ lab//A ┆ null          │
        │ 1          ┆ 2021-01-01 00:00:00 ┆ lab//B ┆ null          │
        │ 2          ┆ 1988-01-02 00:00:00 ┆ DOB    ┆ null          │
        │ 2          ┆ 2023-01-03 00:00:00 ┆ AGE    ┆ 35.004169     │
        │ 2          ┆ 2023-01-03 00:00:00 ┆ lab//A ┆ null          │
        │ 3          ┆ 2022-01-01 00:00:00 ┆ lab//B ┆ null          │
        │ 3          ┆ 2022-01-01 00:00:00 ┆ dx//1  ┆ null          │
        └────────────┴─────────────────────┴────────┴───────────────┘

    We can also specify that some new event codes should go at the end of the event row block, not the
    beginning, via `new_code_last_regex`. This makes the most sense with the `timeline_tokens_fntr`, which
    generates a few types of tokens:

        >>> timeline_tokens_fn = timeline_tokens_fntr(DictConfig({"time_unit": "y"}))
        >>> timeline_tokens_fn(df)
        shape: (8, 4)
        ┌────────────┬─────────────────────┬────────────────────────┬───────────────┐
        │ subject_id ┆ time                ┆ code                   ┆ numeric_value │
        │ ---        ┆ ---                 ┆ ---                    ┆ ---           │
        │ u32        ┆ datetime[μs]        ┆ str                    ┆ f32           │
        ╞════════════╪═════════════════════╪════════════════════════╪═══════════════╡
        │ 1          ┆ 1990-01-01 00:00:00 ┆ TIMELINE//START        ┆ null          │
        │ 1          ┆ 2021-01-01 00:00:00 ┆ TIMELINE//DELTA//years ┆ 31.001347     │
        │ 1          ┆ 2021-01-01 00:00:00 ┆ TIMELINE//END          ┆ null          │
        │ 2          ┆ 1988-01-02 00:00:00 ┆ TIMELINE//START        ┆ null          │
        │ 2          ┆ 2023-01-03 00:00:00 ┆ TIMELINE//DELTA//years ┆ 35.004169     │
        │ 2          ┆ 2023-01-03 00:00:00 ┆ TIMELINE//END          ┆ null          │
        │ 3          ┆ 2022-01-01 00:00:00 ┆ TIMELINE//START        ┆ null          │
        │ 3          ┆ 2022-01-01 00:00:00 ┆ TIMELINE//END          ┆ null          │
        └────────────┴─────────────────────┴────────────────────────┴───────────────┘

    We want to add the TIMELINE//START and TIMELINE//DELTA//* events to the beginning of the event row block,
    but the TIMELINE//END event to the end of the event row block. We can use the `new_code_last_regex`
    parameter to do this:

        >>> add_timeline_tokens_fn = add_new_events_fntr(
        ...     timeline_tokens_fn, new_code_last_regex="TIMELINE//END"
        ... )
        >>> add_timeline_tokens_fn(df)
        shape: (16, 4)
        ┌────────────┬─────────────────────┬────────────────────────┬───────────────┐
        │ subject_id ┆ time                ┆ code                   ┆ numeric_value │
        │ ---        ┆ ---                 ┆ ---                    ┆ ---           │
        │ u32        ┆ datetime[μs]        ┆ str                    ┆ f32           │
        ╞════════════╪═════════════════════╪════════════════════════╪═══════════════╡
        │ 1          ┆ null                ┆ static                 ┆ null          │
        │ 1          ┆ 1990-01-01 00:00:00 ┆ TIMELINE//START        ┆ null          │
        │ 1          ┆ 1990-01-01 00:00:00 ┆ DOB                    ┆ null          │
        │ 1          ┆ 2021-01-01 00:00:00 ┆ TIMELINE//DELTA//years ┆ 31.001347     │
        │ 1          ┆ 2021-01-01 00:00:00 ┆ lab//A                 ┆ null          │
        │ …          ┆ …                   ┆ …                      ┆ …             │
        │ 2          ┆ 2023-01-03 00:00:00 ┆ TIMELINE//END          ┆ null          │
        │ 3          ┆ 2022-01-01 00:00:00 ┆ TIMELINE//START        ┆ null          │
        │ 3          ┆ 2022-01-01 00:00:00 ┆ lab//B                 ┆ null          │
        │ 3          ┆ 2022-01-01 00:00:00 ┆ dx//1                  ┆ null          │
        │ 3          ┆ 2022-01-01 00:00:00 ┆ TIMELINE//END          ┆ null          │
        └────────────┴─────────────────────┴────────────────────────┴───────────────┘
    """

    def out_fn(df: pl.DataFrame) -> pl.DataFrame:
        new_events = fn(df)

        if new_code_last_regex is not None:
            new_events_last = new_events.filter(pl.col("code").str.contains(new_code_last_regex))
            new_events_first = new_events.filter(~pl.col("code").str.contains(new_code_last_regex))

            concat_df = pl.concat([new_events_first, df, new_events_last], how="diagonal")
        else:
            concat_df = pl.concat([new_events, df], how="diagonal")

        return concat_df.sort(by=[DataSchema.subject_id_name, DataSchema.time_name], maintain_order=True)

    return out_fn


@Stage.register
def add_time_derived_measurements(stage_cfg: DictConfig) -> Callable[[pl.LazyFrame], pl.LazyFrame]:
    """Adds all requested time-derived measurements to a DataFrame.

    Args:
        stage_cfg: The configuration for the time-derived measurements. Recognized time derived functors
            include the following keys:
                - "age": The configuration for the age function.
                - "time_of_day": The configuration for the time of day function.
                - "timeline_tokens": The configuration for the timeline tokens function.

    Returns:
        A function that adds all requested time-derived measurements to a DataFrame. A functor can be skipped
        by setting its config to `None`.

    Raises:
        ValueError: If an unrecognized time-derived measurement is requested.

    Examples:
        >>> fn = add_time_derived_measurements({"age": None}) # Does nothing as age is skipped.
        >>> fn(3)
        3
        >>> add_time_derived_measurements(DictConfig({"buzz": {}}))
        Traceback (most recent call last):
            ...
        ValueError: Unknown time-derived measurement: buzz
    """

    map_fns = []
    # We use the raw stages object as the induced `stage_cfg` has extra properties like the input and output
    # directories.
    for feature_name, feature_cfg in stage_cfg.items():
        if feature_cfg is None:
            logger.info(f"Skipping {feature_name} as it is None")
            continue

        match feature_name:
            case "age":
                map_fns.append(add_new_events_fntr(age_fntr(feature_cfg)))
            case "time_of_day":
                map_fns.append(add_new_events_fntr(time_of_day_fntr(feature_cfg)))
            case "timeline_tokens":
                timeline_end_code = feature_cfg.get("timeline_end_code", "TIMELINE//END")
                kwargs = {"new_code_last_regex": timeline_end_code} if timeline_end_code is not None else {}
                map_fns.append(add_new_events_fntr(timeline_tokens_fntr(feature_cfg), **kwargs))
            case str() if feature_name in INFERRED_STAGE_KEYS:
                continue
            case _:
                raise ValueError(f"Unknown time-derived measurement: {feature_name}")

        logger.info(f"Adding {feature_name} via config: {OmegaConf.to_yaml(feature_cfg)}")

    def fn(df: pl.LazyFrame) -> pl.LazyFrame:
        for map_fn in map_fns:
            df = map_fn(df)
        return df

    return fn
